# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: notification_database_data.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()



# pylint: skip-file
DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n notification_database_data.proto\"\xff\t\n\x1dNotificationDatabaseDataProto\x12\"\n\x1apersistent_notification_id\x18\x01 \x01(\x03\x12\x17\n\x0fnotification_id\x18\x05 \x01(\t\x12\x0e\n\x06origin\x18\x02 \x01(\t\x12&\n\x1eservice_worker_registration_id\x18\x03 \x01(\x03\x12&\n\x1ereplaced_existing_notification\x18\x06 \x01(\x08\x12\x12\n\nnum_clicks\x18\x07 \x01(\x05\x12 \n\x18num_action_button_clicks\x18\x08 \x01(\x05\x12\x1c\n\x14\x63reation_time_millis\x18\t \x01(\x03\x12%\n\x1dtime_until_first_click_millis\x18\n \x01(\x03\x12$\n\x1ctime_until_last_click_millis\x18\x0b \x01(\x03\x12\x1f\n\x17time_until_close_millis\x18\x0c \x01(\x03\x12\x42\n\rclosed_reason\x18\r \x01(\x0e\x32+.NotificationDatabaseDataProto.ClosedReason\x12J\n\x11notification_data\x18\x04 \x01(\x0b\x32/.NotificationDatabaseDataProto.NotificationData\x12\x15\n\rhas_triggered\x18\x0e \x01(\x08\x1a\xba\x01\n\x12NotificationAction\x12\x0e\n\x06\x61\x63tion\x18\x01 \x01(\t\x12\r\n\x05title\x18\x02 \x01(\t\x12\x0c\n\x04icon\x18\x03 \x01(\t\x12\x44\n\x04type\x18\x04 \x01(\x0e\x32\x36.NotificationDatabaseDataProto.NotificationAction.Type\x12\x13\n\x0bplaceholder\x18\x05 \x01(\t\"\x1c\n\x04Type\x12\n\n\x06\x42UTTON\x10\x00\x12\x08\n\x04TEXT\x10\x01\x1a\xe4\x03\n\x10NotificationData\x12\r\n\x05title\x18\x01 \x01(\t\x12L\n\tdirection\x18\x02 \x01(\x0e\x32\x39.NotificationDatabaseDataProto.NotificationData.Direction\x12\x0c\n\x04lang\x18\x03 \x01(\t\x12\x0c\n\x04\x62ody\x18\x04 \x01(\t\x12\x0b\n\x03tag\x18\x05 \x01(\t\x12\r\n\x05image\x18\x0f \x01(\t\x12\x0c\n\x04icon\x18\x06 \x01(\t\x12\r\n\x05\x62\x61\x64ge\x18\x0e \x01(\t\x12\x1d\n\x11vibration_pattern\x18\t \x03(\x05\x42\x02\x10\x01\x12\x11\n\ttimestamp\x18\x0c \x01(\x03\x12\x10\n\x08renotify\x18\r \x01(\x08\x12\x0e\n\x06silent\x18\x07 \x01(\x08\x12\x1b\n\x13require_interaction\x18\x0b \x01(\x08\x12\x0c\n\x04\x64\x61ta\x18\x08 \x01(\x0c\x12\x42\n\x07\x61\x63tions\x18\n \x03(\x0b\x32\x31.NotificationDatabaseDataProto.NotificationAction\x12\x1e\n\x16show_trigger_timestamp\x18\x10 \x01(\x03\";\n\tDirection\x12\x11\n\rLEFT_TO_RIGHT\x10\x00\x12\x11\n\rRIGHT_TO_LEFT\x10\x01\x12\x08\n\x04\x41UTO\x10\x02\"4\n\x0c\x43losedReason\x12\x08\n\x04USER\x10\x00\x12\r\n\tDEVELOPER\x10\x01\x12\x0b\n\x07UNKNOWN\x10\x02')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'notification_database_data_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_NOTIFICATIONDATABASEDATAPROTO_NOTIFICATIONDATA'].fields_by_name['vibration_pattern']._options = None
  _globals['_NOTIFICATIONDATABASEDATAPROTO_NOTIFICATIONDATA'].fields_by_name['vibration_pattern']._serialized_options = b'\020\001'
  _globals['_NOTIFICATIONDATABASEDATAPROTO']._serialized_start=37
  _globals['_NOTIFICATIONDATABASEDATAPROTO']._serialized_end=1316
  _globals['_NOTIFICATIONDATABASEDATAPROTO_NOTIFICATIONACTION']._serialized_start=589
  _globals['_NOTIFICATIONDATABASEDATAPROTO_NOTIFICATIONACTION']._serialized_end=775
  _globals['_NOTIFICATIONDATABASEDATAPROTO_NOTIFICATIONACTION_TYPE']._serialized_start=747
  _globals['_NOTIFICATIONDATABASEDATAPROTO_NOTIFICATIONACTION_TYPE']._serialized_end=775
  _globals['_NOTIFICATIONDATABASEDATAPROTO_NOTIFICATIONDATA']._serialized_start=778
  _globals['_NOTIFICATIONDATABASEDATAPROTO_NOTIFICATIONDATA']._serialized_end=1262
  _globals['_NOTIFICATIONDATABASEDATAPROTO_NOTIFICATIONDATA_DIRECTION']._serialized_start=1203
  _globals['_NOTIFICATIONDATABASEDATAPROTO_NOTIFICATIONDATA_DIRECTION']._serialized_end=1262
  _globals['_NOTIFICATIONDATABASEDATAPROTO_CLOSEDREASON']._serialized_start=1264
  _globals['_NOTIFICATIONDATABASEDATAPROTO_CLOSEDREASON']._serialized_end=1316
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Parser plugin for Chrome Notifications."""
from __future__ import annotations

import dataclasses
import logging
from typing import TYPE_CHECKING, Optional

try:
  # pytype: disable=import-error
  from dfdatetime import webkit_time

  from dfindexeddb.leveldb.plugins import notification_database_data_pb2 as notification_pb2

  # pytype: enable=import-error
  _has_import_dependencies = True
except ImportError as err:
  _has_import_dependencies = False
  logging.warning(
      (
          "Could not import dependencies for "
          "leveldb.plugins.chrome_notifications: %s"
      ),
      err,
  )

from dfindexeddb.indexeddb.chromium import blink
from dfindexeddb.leveldb.plugins import interface, manager

if TYPE_CHECKING:
  from dfindexeddb.leveldb import ldb, log


@dataclasses.dataclass
class ChromeNotificationRecord(interface.LeveldbPlugin):
  """Chrome notification record."""

  src_file: Optional[str] = None
  offset: Optional[int] = None
  key: Optional[str] = None
  sequence_number: Optional[int] = None
  type: Optional[int] = None
  origin: Optional[str] = None
  service_worker_registration_id: Optional[int] = None
  notification_title: Optional[str] = None
  notification_direction: Optional[str] = None
  notification_lang: Optional[str] = None
  notification_body: Optional[str] = None
  notification_tag: Optional[str] = None
  notification_icon: Optional[str] = None
  notification_silent: Optional[bool] = None
  notification_data: Optional[str] = None
  notification_require_interaction: Optional[bool] = None
  notification_time: Optional[str] = None
  notification_renotify: Optional[bool] = None
  notification_badge: Optional[str] = None
  notification_image: Optional[str] = None
  notification_id: Optional[str] = None
  replaced_existing_notification: Optional[bool] = None
  num_clicks: Optional[int] = None
  num_action_button_clicks: Optional[int] = None
  creation_time: Optional[str] = None
  closed_reason: Optional[str] = None
  has_triggered: Optional[bool] = None

  @classmethod
  def FromKeyValueRecord(
      cls, ldb_record: ldb.KeyValueRecord | log.ParsedInternalKey
  ) -> ChromeNotificationRecord:
    record = cls()
    record.offset = ldb_record.offset
    record.key = ldb_record.key.decode()
    record.sequence_number = ldb_record.sequence_number
    record.type = ldb_record.record_type

    if not ldb_record.value:
      return record

    # pylint: disable-next=no-member,line-too-long
    notification_proto = notification_pb2.NotificationDatabaseDataProto()  # type: ignore[attr-defined]
    notification_proto.ParseFromString(ldb_record.value)

    record.origin = notification_proto.origin
    record.service_worker_registration_id = (
        notification_proto.service_worker_registration_id
    )
    record.notification_title = notification_proto.notification_data.title
    record.notification_direction = (
        notification_proto.notification_data.direction
    )
    record.notification_lang = notification_proto.notification_data.lang
    record.notification_body = notification_proto.notification_data.body
    record.notification_tag = notification_proto.notification_data.tag
    record.notification_icon = notification_proto.notification_data.icon
    record.notification_silent = notification_proto.notification_data.silent
    record.notification_data = notification_proto.notification_data.data
    record.notification_require_interaction = (
        notification_proto.notification_data.require_interaction
    )
    record.notification_time = webkit_time.WebKitTime(
        timestamp=notification_proto.notification_data.timestamp
    ).CopyToDateTimeString()
    record.notification_renotify = notification_proto.notification_data.renotify
    record.notification_badge = notification_proto.notification_data.badge
    record.notification_image = notification_proto.notification_data.image
    record.notification_id = notification_proto.notification_id
    record.replaced_existing_notification = (
        notification_proto.replaced_existing_notification
    )
    record.num_clicks = notification_proto.num_clicks
    record.num_action_button_clicks = (
        notification_proto.num_action_button_clicks
    )
    record.creation_time = webkit_time.WebKitTime(
        timestamp=notification_proto.creation_time_millis
    ).CopyToDateTimeString()
    record.closed_reason = notification_proto.closed_reason
    record.has_triggered = notification_proto.has_triggered

    if not notification_proto.notification_data.data:
      return record

    notification_data = blink.V8ScriptValueDecoder(
        raw_data=notification_proto.notification_data.data
    ).Deserialize()
    record.notification_data = notification_data

    return record


# check if dependencies are in existence..

if _has_import_dependencies:
  manager.PluginManager.RegisterPlugin(ChromeNotificationRecord)



# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Parser for LevelDB Table (.ldb) files."""
from __future__ import annotations

import io
import os
from dataclasses import dataclass, field
from typing import BinaryIO, Iterable, Tuple

import snappy
import zstd

from dfindexeddb.leveldb import definitions, utils


@dataclass
class KeyValueRecord:
  """A leveldb table key-value record.

  Attributes:
    offset: the offset of the record.
    key: the key of the record.
    value: the value of the record.
    sequence_number: the sequence number of the record.
    record_type: the type of the record.
  """

  offset: int
  key: bytes
  value: bytes
  sequence_number: int
  record_type: definitions.InternalRecordType

  @classmethod
  def FromDecoder(
      cls, decoder: utils.LevelDBDecoder, block_offset: int, shared_key: bytes
  ) -> Tuple[KeyValueRecord, bytes]:
    """Decodes a ldb key value record.

    Args:
      decoder: the leveldb decoder.
      block_offset: the block offset.
      shared_key: the shared key bytes.

    Returns:
      A tuple of the parsed KeyValueRecord and the updated shared key bytes.
    """
    offset, shared_bytes = decoder.DecodeUint32Varint()
    _, unshared_bytes = decoder.DecodeUint32Varint()
    _, value_length = decoder.DecodeUint32Varint()
    _, key_delta = decoder.ReadBytes(unshared_bytes)
    _, value = decoder.ReadBytes(value_length)

    shared_key = shared_key[:shared_bytes] + key_delta
    key = shared_key[: -definitions.PACKED_SEQUENCE_AND_TYPE_LENGTH]
    sequence_number = int.from_bytes(
        shared_key[-definitions.SEQUENCE_LENGTH :],
        byteorder="little",
        signed=False,
    )
    key_type = shared_key[-definitions.PACKED_SEQUENCE_AND_TYPE_LENGTH]
    record_type = definitions.InternalRecordType(key_type)
    return (
        cls(
            offset=offset + block_offset,
            key=key,
            value=value,
            sequence_number=sequence_number,
            record_type=record_type,
        ),
        shared_key,
    )


@dataclass
class Block:
  """A leveldb table block.

  Attributes:
    offset: the offset of the block.
    block_offset:
  """

  offset: int
  block_offset: int
  length: int
  data: bytes = field(repr=False)
  footer: bytes  # 5 bytes = 1 byte compressed flag + 4 bytes checksum.

  def IsSnappyCompressed(self) -> bool:
    """Returns true if the block is snappy compressed."""
    return self.footer[0] == definitions.BlockCompressionType.SNAPPY

  def IsZstdCompressed(self) -> bool:
    """Returns true if the block is zstd compressed."""
    return self.footer[0] == definitions.BlockCompressionType.ZSTD

  def GetBuffer(self) -> bytes:
    """Returns the block buffer, decompressing if required."""
    if self.IsSnappyCompressed():
      return bytes(snappy.decompress(self.data))
    if self.IsZstdCompressed():
      return bytes(zstd.decompress(self.data))
    return self.data

  def GetRecords(self) -> Iterable[KeyValueRecord]:
    """Returns an iterator over the key value records in the block.

    Yields:
      KeyValueRecords
    """
    # get underlying block content, decompressing if required
    buffer = self.GetBuffer()
    decoder = utils.LevelDBDecoder(io.BytesIO(buffer))

    # trailer of a block has the form:
    #    restarts: uint32[num_restarts]
    #    num_restarts: uint32
    decoder.stream.seek(-definitions.BLOCK_RESTART_ENTRY_LENGTH, os.SEEK_END)
    _, num_restarts = decoder.DecodeUint32()
    restarts_offset = (decoder.stream.tell()) - (
        (num_restarts + 1) * definitions.BLOCK_RESTART_ENTRY_LENGTH
    )

    decoder.stream.seek(restarts_offset)
    _, offset = decoder.DecodeUint32()
    decoder.stream.seek(offset)
    key = b""

    while decoder.stream.tell() < restarts_offset:
      key_value_record, key = KeyValueRecord.FromDecoder(
          decoder, self.block_offset, key
      )
      yield key_value_record

    # TODO: parse trailer of the block for restart points (where the full
    # key is stored to allow for binary lookup).  It's not needed at this time
    # since we are sequentially iterating over the records in the block/file.


@dataclass
class BlockHandle(utils.FromDecoderMixin):
  """A handle to a block in the ldb file.

  Attributes:
    offset: the offset of the block handle.
    block_offset: the offset of the block.
    length: the length of the block.
  """

  offset: int
  block_offset: int
  length: int

  def Load(self, stream: BinaryIO) -> Block:
    """Loads the block data.

    Args:
      stream: the binary stream of the ldb file.

    Returns:
      a Block.

    Raises:
      ValueError: if it could not read all of the block or block footer.
    """
    stream.seek(self.block_offset, os.SEEK_SET)
    data = stream.read(self.length)
    if len(data) != self.length:
      raise ValueError("Could not read all of the block")

    footer = stream.read(definitions.BLOCK_TRAILER_SIZE)
    if len(footer) != definitions.BLOCK_TRAILER_SIZE:
      raise ValueError("Could not read all of the block footer")

    return Block(self.offset, self.block_offset, self.length, data, footer)

  @classmethod
  def FromDecoder(
      cls, decoder: utils.LevelDBDecoder, base_offset: int = 0
  ) -> BlockHandle:
    """Decodes a BlockHandle from the current position of a LevelDBDecoder.

    Args:
      decoder: the LevelDBDecoder.
      base_offset: the base offset.

    Returns:
      The BlockHandle instance.
    """
    offset, block_offset = decoder.DecodeUint64Varint()
    _, length = decoder.DecodeUint64Varint()
    return cls(offset + base_offset, block_offset, length)


class FileReader:
  """A leveldb table (.ldb or .sst) file reader.

  A Ldb FileReader provides read-only sequential iteration of serialized
  structures in a leveldb ldb file.  These structures include:
  * blocks (Block)
  * records (KeyValueRecord)
  """

  def __init__(self, filename: str):
    """Initializes the LogFile.

    Args:
      filename: the .ldb filename.

    Raises:
      ValueError if the file has an invalid magic number at the end.
    """
    self.filename = filename
    with open(self.filename, "rb") as fh:
      fh.seek(-len(definitions.TABLE_MAGIC), os.SEEK_END)
      if fh.read(len(definitions.TABLE_MAGIC)) != definitions.TABLE_MAGIC:
        raise ValueError(f"Invalid magic number in {self.filename}")

      fh.seek(-definitions.TABLE_FOOTER_SIZE, os.SEEK_END)
      # meta_handle, need to read first due to variable integers
      _ = BlockHandle.FromStream(fh)
      index_handle = BlockHandle.FromStream(fh)

      # self.meta_block = meta_handle.load(fh)  # TODO: support meta blocks
      self.index_block = index_handle.Load(fh)

  def GetBlocks(self) -> Iterable[Block]:
    """Returns an iterator of Blocks.

    Yields:
      Block.
    """
    with open(self.filename, "rb") as fh:
      for key_value_record in self.index_block.GetRecords():
        block_handle = BlockHandle.FromStream(
            io.BytesIO(key_value_record.value),
            base_offset=key_value_record.offset,
        )
        yield block_handle.Load(fh)

  def GetKeyValueRecords(self) -> Iterable[KeyValueRecord]:
    """Returns an iterator of KeyValueRecords.

    Yields:
      KeyValueRecords.
    """
    for block in self.GetBlocks():
      yield from block.GetRecords()

  def RangeIter(self) -> Iterable[Tuple[bytes, bytes]]:  # pylint: disable=C0103
    """Returns an iterator of key-value pairs.

    Yields:
      A tuple of key and value as bytes.
    """
    for record in self.GetKeyValueRecords():
      yield record.key, record.value



# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Definitions for IndexedDB."""
from enum import Enum, IntEnum, IntFlag

REQUIRES_PROCESSING_SSV_PSEUDO_VERSION = 0x11
REPLACE_WITH_BLOB = 0x01
COMPRESSED_WITH_SNAPPY = 0x02


class DatabaseMetaDataKeyType(IntEnum):
  """Database Metadata key types."""

  ORIGIN_NAME = 0
  DATABASE_NAME = 1
  IDB_STRING_VERSION_DATA = 2
  MAX_ALLOCATED_OBJECT_STORE_ID = 3
  IDB_INTEGER_VERSION = 4
  BLOB_NUMBER_GENERATOR_CURRENT_NUMBER = 5
  OBJECT_STORE_META_DATA = 50
  INDEX_META_DATA = 100
  OBJECT_STORE_FREE_LIST = 150
  INDEX_FREE_LIST = 151
  OBJECT_STORE_NAMES = 200
  INDEX_NAMES = 201


class GlobalMetadataKeyType(IntEnum):
  """Global Metadata key types."""

  SCHEMA_VERSION = 0
  MAX_DATABASE_ID = 1
  DATA_VERSION = 2
  RECOVERY_BLOB_JOURNAL = 3
  ACTIVE_BLOB_JOURNAL = 4
  EARLIEST_SWEEP = 5
  EARLIEST_COMPACTION_TIME = 6
  SCOPES_PREFIX = 50
  DATABASE_FREE_LIST = 100
  DATABASE_NAME = 201


class IDBKeyPathType(IntEnum):
  """IndexedDB key path types."""

  NULL = 0
  STRING = 1
  ARRAY = 2


class IDBKeyType(IntEnum):
  """IndexedDB key types."""

  NULL = 0
  STRING = 1
  DATE = 2
  NUMBER = 3
  ARRAY = 4
  MIN_KEY = 5
  BINARY = 6


class IndexMetaDataKeyType(IntEnum):
  """IndexedDB metadata key types."""

  INDEX_NAME = 0
  UNIQUE_FLAG = 1
  KEY_PATH = 2
  MULTI_ENTRY_FLAG = 3


class KeyPrefixType(Enum):
  """IndexedDB key prefix types."""

  GLOBAL_METADATA = 0
  DATABASE_METADATA = 1
  OBJECT_STORE_DATA = 2
  EXISTS_ENTRY = 3
  INDEX_DATA = 4
  INVALID_TYPE = 5
  BLOB_ENTRY = 6


class ObjectStoreMetaDataKeyType(IntEnum):
  """IndexedDB object store metadata key types."""

  OBJECT_STORE_NAME = 0
  KEY_PATH = 1
  AUTO_INCREMENT_FLAG = 2
  IS_EVICTABLE = 3
  LAST_VERSION_NUMBER = 4
  MAXIMUM_ALLOCATED_INDEX_ID = 5
  HAS_KEY_PATH = 6
  KEY_GENERATOR_CURRENT_NUMBER = 7


class ExternalObjectType(IntEnum):
  """IndexedDB external object types."""

  BLOB = 0
  FILE = 1
  FILE_SYSTEM_ACCESS_HANDLE = 2


class BlinkSerializationTag(IntEnum):
  """Blink Javascript serialization tags."""

  MESSAGE_PORT = ord("M")
  MOJO_HANDLE = ord("h")
  BLOB = ord("b")
  BLOB_INDEX = ord("i")
  FILE = ord("f")
  FILE_INDEX = ord("e")
  DOM_FILE_SYSTEM = ord("d")
  FILE_SYSTEM_FILE_HANDLE = ord("n")
  FILE_SYSTEM_DIRECTORY_HANDLE = ord("N")
  FILE_LIST = ord("l")
  FILE_LIST_INDEX = ord("L")
  IMAGE_DATA = ord("#")
  IMAGE_BITMAP = ord("g")
  IMAGE_BITMAP_TRANSFER = ord("G")
  OFFSCREEN_CANVAS_TRANSFER = ord("H")
  READABLE_STREAM_TRANSFER = ord("r")
  TRANSFORM_STREAM_TRANSFER = ord("m")
  WRITABLE_STREAM_TRANSFER = ord("w")
  MEDIA_STREAM_TRACK = ord("s")
  DOM_POINT = ord("Q")
  DOM_POINT_READ_ONLY = ord("W")
  DOM_RECT = ord("E")
  DOM_RECT_READ_ONLY = ord("R")
  DOM_QUAD = ord("T")
  DOM_MATRIX = ord("Y")
  DOM_MATRIX_READ_ONLY = ord("U")
  DOM_MATRIX2D = ord("I")
  DOM_MATRIX2D_READ_ONLY = ord("O")
  CRYPTO_KEY = ord("K")
  RTC_CERTIFICATE = ord("k")
  RTC_ENCODED_AUDIO_FRAME = ord("A")
  RTC_ENCODED_VIDEO_FRAME = ord("V")
  AUDIO_DATA = ord("a")
  VIDEO_FRAME = ord("v")
  ENCODED_AUDIO_CHUNK = ord("y")
  ENCODED_VIDEO_CHUNK = ord("z")
  CROP_TARGET = ord("c")
  RESTRICTION_TARGET = ord("D")
  MEDIA_SOURCE_HANDLE = ord("S")
  DEPRECATED_DETECTED_BARCODE = ord("B")
  DEPRECATED_DETECTED_FACE = ord("F")
  DEPRECATED_DETECTED_TEXT = ord("t")
  FENCED_FRAME_CONFIG = ord("C")
  DOM_EXCEPTION = ord("x")
  TRAILER_OFFSET = 0xFE
  VERSION = 0xFF
  TRAILER_REQUIRES_INTERFACES = 0xA0


class CryptoKeyAlgorithm(IntEnum):
  """CryptoKey Algorithm types."""

  AES_CBC = 1
  HMAC = 2
  RSA_SSA_PKCS1_V1_5 = 3
  SHA1 = 5
  SHA256 = 6
  SHA384 = 7
  SHA512 = 8
  AES_GCM = 9
  RSA_OAEP = 10
  AES_CTR = 11
  AES_KW = 12
  RSA_PSS = 13
  ECDSA = 14
  ECDH = 15
  HKDF = 16
  PBKDF2 = 17
  ED25519 = 18


class NamedCurve(IntEnum):
  """Named Curve types."""

  P256 = 1
  P384 = 2
  P521 = 3


class CryptoKeyUsage(IntFlag):
  """CryptoKey Usage flags."""

  EXTRACTABLE = 1 << 0
  ENCRYPT = 1 << 1
  DECRYPT = 1 << 2
  SIGN = 1 << 3
  VERIFY = 1 << 4
  DERIVE_KEY = 1 << 5
  WRAP_KEY = 1 << 6
  UNWRAP_KEY = 1 << 7
  DRIVE_BITS = 1 << 8


class CryptoKeySubTag(IntEnum):
  """CryptoKey subtag types."""

  AES_KEY = 1
  HMAC_KEY = 2
  RSA_HASHED_KEY = 4
  EC_KEY = 5
  NO_PARAMS_KEY = 6
  ED25519_KEY = 7


class AsymmetricCryptoKeyType(IntEnum):
  """Asymmetric CryptoKey types."""

  PUBLIC_KEY = 1
  PRIVATE_KEY = 2


class WebCryptoKeyType(Enum):
  """WebCryptoKey types."""

  SECRET = "Secret"
  PUBLIC = "Public"
  PRIVATE = "Private"


class V8SerializationTag(IntEnum):
  """V8 Javascript serialization tags."""

  VERSION = 0xFF
  PADDING = ord("\0")
  VERIFY_OBJECT_COUNT = ord("?")
  THE_HOLE = ord("-")
  UNDEFINED = ord("_")
  NULL = ord("0")
  TRUE = ord("T")
  FALSE = ord("F")
  INT32 = ord("I")
  UINT32 = ord("U")
  DOUBLE = ord("N")
  BIGINT = ord("Z")
  UTF8_STRING = ord("S")
  ONE_BYTE_STRING = ord('"')
  TWO_BYTE_STRING = ord("c")
  OBJECT_REFERENCE = ord("^")
  BEGIN_JS_OBJECT = ord("o")
  END_JS_OBJECT = ord("{")
  BEGIN_SPARSE_JS_ARRAY = ord("a")
  END_SPARSE_JS_ARRAY = ord("@")
  BEGIN_DENSE_JS_ARRAY = ord("A")
  END_DENSE_JS_ARRAY = ord("$")
  DATE = ord("D")
  TRUE_OBJECT = ord("y")
  FALSE_OBJECT = ord("x")
  NUMBER_OBJECT = ord("n")
  BIGINT_OBJECT = ord("z")
  STRING_OBJECT = ord("s")
  REGEXP = ord("R")
  BEGIN_JS_MAP = ord(";")
  END_JS_MAP = ord(":")
  BEGIN_JS_SET = ord("'")
  END_JS_SET = ord(",")
  ARRAY_BUFFER = ord("B")
  RESIZABLE_ARRAY_BUFFER = ord("~")
  ARRAY_BUFFER_TRANSFER = ord("t")
  ARRAY_BUFFER_VIEW = ord("V")
  SHARED_ARRAY_BUFFER = ord("u")
  SHARED_OBJECT = ord("p")
  WASM_MODULE_TRANSFER = ord("w")
  HOST_OBJECT = ord("\\")
  WASM_MEMORY_TRANSFER = ord("m")
  ERROR = ord("r")
  LEGACY_RESERVED_MESSAGE_PORT = ord("M")
  LEGACY_RESERVED_BLOB = ord("b")
  LEGACY_RESERVED_BLOB_INDEX = ord("i")
  LEGACY_RESERVED_FILE = ord("f")
  LEGACY_RESERVED_FILE_INDEX = ord("e")
  LEGACY_RESERVED_DOM_FILE_SYSTEM = ord("d")
  LEGACY_RESERVED_FILE_LIST = ord("l")
  LEGACY_RESERVED_FILE_LIST_INDEX = ord("L")
  LEGACY_RESERVED_IMAGE_DATA = ord("#")
  LEGACY_RESERVED_IMAGE_BITMAP = ord("g")
  LEGACY_RESERVED_IMAGE_BITMAP_TRANSFER = ord("G")
  LEGACY_RESERVED_OFF_SCREEN_CANVAS = ord("H")
  LEGACY_RESERVED_CRYPTO_KEY = ord("K")
  LEGACY_RESERVED_RTC_CERTIFICATE = ord("k")


class V8ArrayBufferViewTag(IntEnum):
  """V8 ArrayBufferView tags."""

  INT8_ARRAY = ord("b")
  UINT8_ARRAY = ord("B")
  UINT8_CLAMPED_ARRAY = ord("C")
  INT16_ARRAY = ord("w")
  UINT16_ARRAY = ord("W")
  INT32_ARRAY = ord("d")
  UINT32_ARRAY = ord("D")
  FLOAT32_ARRAY = ord("f")
  FLOAT64_ARRAY = ord("F")
  BIGINT64_ARRAY = ord("q")
  BIGUINT64_ARRAY = ord("Q")
  DATAVIEW = ord("?")


class V8ErrorTag(IntEnum):
  """V8 Error tags."""

  EVAL_ERROR_PROTOTYPE = ord("E")
  RANGE_ERROR_PROTOTYPE = ord("R")
  REFERENCE_ERROR_PROTOTYPE = ord("F")
  SYNTAX_ERROR_PROTOTYPE = ord("S")
  TYPE_ERROR_PROTOTYPE = ord("T")
  URI_ERROR_PROTOTYPE = ord("U")
  MESSAGE = ord("m")
  CAUSE = ord("c")
  STACK = ord("s")
  END = ord(".")


class ImageSerializationTag(IntEnum):
  """Image Serialization tags."""

  END = 0
  PREDEFINED_COLOR_SPACE = 1
  CANVAS_PIXEL_FORMAT = 2
  IMAGE_DATA_STORAGE_FORMAT = 3
  ORIGIN_CLEAN = 4
  IS_PREMULTIPLIED = 5
  CANVAS_OPACITY_MODE = 6
  PARAMETRIC_COLOR_SPACE = 7
  IMAGE_ORIENTATION = 8
  LAST = IMAGE_ORIENTATION


class SerializedPredefinedColorSpace(IntEnum):
  """Serialized Predefined Color Space enumeration."""

  LEGACY_OBSOLETE = 0
  SRGB = 1
  REC2020 = 2
  P3 = 3
  REC2100HLG = 4
  REC2100PQ = 5
  SRGB_LINEAR = 6
  LAST = SRGB_LINEAR


class SerializedPixelFormat(IntEnum):
  """Serialized Pixel Format enumeration."""

  NATIVE8_LEGACY_OBSOLETE = 0
  F16 = 1
  RGBA8 = 2
  BGRA8 = 3
  RGBX8 = 4
  LAST = RGBX8


class SerializedImageDataStorageFormat(IntEnum):
  """The Serialized Image Data Storage Format."""

  UINT8CLAMPED = 0
  UINT16 = 1
  FLOAT32 = 2
  LAST = FLOAT32


class SerializedOpacityMode(IntEnum):
  """The Serialized Opacity Mode."""

  KNONOPAQUE = 0
  KOPAQUE = 1
  KLAST = KOPAQUE


class SerializedImageOrientation(IntEnum):
  """The Serialized Image Orientation."""

  TOP_LEFT = 0
  TOP_RIGHT = 1
  BOTTOM_RIGHT = 2
  BOTTOM_LEFT = 3
  LEFT_TOP = 4
  RIGHT_TOP = 5
  RIGHT_BOTTOM = 6
  LEFT_BOTTOM = 7
  LAST = LEFT_BOTTOM



# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Firefox IndexedDB records."""
import pathlib
import sqlite3
import sys
import traceback
from dataclasses import dataclass
from typing import Any, Generator, Optional

from dfindexeddb import errors
from dfindexeddb.indexeddb.firefox import gecko


@dataclass
class FirefoxObjectStoreInfo:
  """A FireFox ObjectStoreInfo.

  Attributes:
    id: the object store ID.
    name: the object store name.
    key_path: the object store key path.
    auto_inc: the current auto-increment value.
    database_name: the database name from the database table.
  """

  id: int
  name: str
  key_path: str
  auto_inc: int
  database_name: str


@dataclass
class FirefoxIndexedDBRecord:
  """A Firefox IndexedDBRecord.

  Attributes:
    key: the parsed key.
    value: the parsed value.
    file_ids: the file identifiers.
    object_store_id: the object store id.
    object_store_name: the object store name from the object_store table.
    database_name: the IndexedDB database name from the database table.
  """

  key: Any
  value: Any
  file_ids: Optional[str]
  object_store_id: int
  object_store_name: str
  database_name: str


class FileReader:
  """A reader for Firefox IndexedDB sqlite3 files.

  Attributes:
    database_name: the database name.
    origin: the database origin.
    metadata_version: the metadata version.
    last_vacuum_time: the last vacuum time.
    last_analyze_time: the last analyze time.
  """

  def __init__(self, filename: str):
    """Initializes the FileReader.

    Args:
      filename: the IndexedDB filename.
    """
    self.filename = filename

    with sqlite3.connect(f"file:{self.filename}?mode=ro", uri=True) as conn:
      cursor = conn.execute(
          "SELECT name, origin, version, last_vacuum_time, last_analyze_time "
          "FROM database"
      )
      result = cursor.fetchone()
      self.database_name = result[0]
      self.origin = result[1]
      self.metadata_version = result[2]
      self.last_vacuum_time = result[3]
      self.last_analyze_time = result[4]

  def _ParseKey(self, key: bytes) -> Any:
    """Parses a key."""
    try:
      return gecko.IDBKey.FromBytes(key)
    except errors.ParserError as e:
      print("failed to parse", key, file=sys.stderr)
      traceback.print_exception(type(e), e, e.__traceback__)
      return key

  def _ParseValue(self, value: bytes) -> Any:
    """Parses a value."""
    try:
      return gecko.JSStructuredCloneDecoder.FromBytes(value)
    except errors.ParserError as err:
      print("failed to parse", value, file=sys.stderr)
      traceback.print_exception(type(err), err, err.__traceback__)
      return value

  def ObjectStores(self) -> Generator[FirefoxObjectStoreInfo, None, None]:
    """Returns the Object Store information from the IndexedDB database.

    Yields:
      FirefoxObjectStoreInfo instances.
    """
    with sqlite3.connect(f"file:{self.filename}?mode=ro", uri=True) as conn:
      cursor = conn.execute(
          "SELECT id, auto_increment, name, key_path FROM object_store"
      )
      results = cursor.fetchall()
      for result in results:
        yield FirefoxObjectStoreInfo(
            id=result[0],
            name=result[2],
            key_path=result[3],
            auto_inc=result[1],
            database_name=self.database_name,
        )

  def RecordsByObjectStoreId(
      self, object_store_id: int
  ) -> Generator[FirefoxIndexedDBRecord, None, None]:
    """Returns FirefoxIndexedDBRecords by a given object store id.

    Args:
      object_store_id: the object store id.
    """
    with sqlite3.connect(f"file:{self.filename}?mode=ro", uri=True) as conn:
      conn.text_factory = bytes
      cursor = conn.execute(
          "SELECT od.key, od.data, od.object_store_id, od.file_ids, os.name "
          "FROM object_data od "
          "JOIN object_store os ON od.object_store_id == os.id "
          "WHERE os.id = ? ORDER BY od.key",
          (object_store_id,),
      )
      for row in cursor:
        key = self._ParseKey(row[0])
        if row[3]:
          value = row[1]
        else:
          value = self._ParseValue(row[1])
        yield FirefoxIndexedDBRecord(
            key=key,
            value=value,
            object_store_id=row[2],
            file_ids=row[3],
            object_store_name=row[4].decode("utf-8"),
            database_name=self.database_name,
        )

  def Records(self) -> Generator[FirefoxIndexedDBRecord, None, None]:
    """Returns FirefoxIndexedDBRecords from the database."""
    with sqlite3.connect(f"file:{self.filename}?mode=ro", uri=True) as conn:
      conn.text_factory = bytes
      cursor = conn.execute(
          "SELECT od.key, od.data, od.object_store_id, od.file_ids, os.name "
          "FROM object_data od "
          "JOIN object_store os ON od.object_store_id == os.id"
      )
      for row in cursor:
        key = self._ParseKey(row[0])
        if row[3]:
          value = row[1]
        else:
          value = self._ParseValue(row[1])
        yield FirefoxIndexedDBRecord(
            key=key,
            value=value,
            object_store_id=row[2],
            file_ids=row[3],
            object_store_name=row[4].decode("utf-8"),
            database_name=self.database_name,
        )


class FolderReader:
  """A reader for a FireFox IndexedDB folder.

  The path takes a general form of ./<origin>/idb/<filename>.[files|sqlite]

  """

  def __init__(self, folder_name: pathlib.Path):
    """Initializes the FireFox IndexedDB FolderReader.

    Args:
      folder_name: the IndexedDB folder name (the origin folder).

    Raises:
      ValueError: if the folder does not exist or is not a directory.
    """
    self.folder_name = folder_name
    if not self.folder_name.exists():
      raise ValueError(f"{folder_name} does not exist.")
    if not self.folder_name.is_dir():
      raise ValueError(f"{folder_name} is not a directory.")

    self.file_names: list[pathlib.Path] = []
    for file_name in self.folder_name.rglob("idb/*.sqlite"):
      self.file_names.append(file_name)

  def Records(self) -> Generator[FirefoxIndexedDBRecord, None, None]:
    """Returns FirefoxIndexedDBRecords from the IndexedDB folder."""
    for file_name in self.file_names:
      yield from FileReader(str(file_name)).Records()



# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Parsers for v8 javascript serialized objects."""
from __future__ import annotations

import io
import os
from dataclasses import dataclass
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Any,
    BinaryIO,
    Dict,
    Optional,
    Set,
    Tuple,
    Union,
)

from dfindexeddb import errors, utils
from dfindexeddb.indexeddb import types
from dfindexeddb.indexeddb.chromium import definitions

if TYPE_CHECKING:
  from dfindexeddb.indexeddb.chromium import blink


@dataclass
class ArrayBufferView:
  """A parsed Javascript ArrayBufferView."""

  buffer: bytes
  tag: definitions.V8ArrayBufferViewTag
  offset: int
  length: int
  flags: int


class ValueDeserializer:
  """A class to deserialize v8 Javascript values/objects.

  Attributes:
    decoder (utils.StreamDecoder): A StreamDecoder that wraps the binary
      stream containing the serialized Javascript value/object.
    delegate: An instance to delegate deserializing an object
    next_id (int): the next ID to use for a deserialized object.
    objects (dict[int, Any]): a dictionary mapping integer IDs to a
      deserialized object.
    version (int): the version of the serialized Javascript value/object
  """

  LATEST_VERSION = 15

  def __init__(self, stream: BinaryIO, delegate: blink.V8ScriptValueDecoder):
    """Initializes a ValueDeserializer.

    Args:
      stream: a stream of bytes to be deserialized.
      delegate: an object to delegate additional deserialization of additional
        Javascript types/objects.
    """
    self.decoder = utils.StreamDecoder(stream)
    self.delegate = delegate
    self.next_id = 0
    self.objects: dict[int, Any] = {}
    self.version = 0

  def GetWireFormatVersion(self) -> int:
    """Returns the underlying wire format version.

    Note: Return value is only valid after a call to ReadHeader.
    """
    return self.version

  def ReadHeader(self) -> bool:
    """Reads the header and returns True if the version is supported."""
    if self._ReadTag() != definitions.V8SerializationTag.VERSION:
      return False
    _, self.version = self.decoder.DecodeUint32Varint()
    if self.version > self.LATEST_VERSION:
      return False
    return True

  def ReadObjectWrapper(self) -> Any:
    """Deserializes a V8 object from the current decoder position."""
    original_position = self.decoder.stream.tell()
    result = self._ReadObject()
    if result is None and self.version == 13:
      self.decoder.stream.seek(original_position, os.SEEK_SET)
      result = self._ReadObject()
      raise NotImplementedError("ValueDeserializer.ReadObjectWrapper v13")
    return result

  def ReadObjectUsingEntireBufferForLegacyFormat(self) -> Any:
    """Reads an object, consuming the entire buffer."""
    raise NotImplementedError(
        "ValueDeserializer.ReadObjectUsingEntireBufferForLegacyFormat"
    )

  def ReadValue(self) -> Any:
    """Reads the Javascript value."""
    if self.GetWireFormatVersion() > 0:
      return self.ReadObjectWrapper()
    return self.ReadObjectUsingEntireBufferForLegacyFormat()

  def _PeekTag(self) -> Optional[definitions.V8SerializationTag]:
    """Peeks the next serialization tag from the current decoder position.

    Returns:
      The serialization tag or None if there is no more bytes to read.
    """
    try:
      _, tag_value = self.decoder.PeekBytes(1)
    except errors.DecoderError:
      return None
    try:
      return definitions.V8SerializationTag(tag_value[0])
    except ValueError as error:
      raise errors.ParserError(
          f"Invalid v8 tag value {tag_value!r} at offset"
          f" {self.decoder.stream.tell()}"
      ) from error

  def _ReadTag(self) -> definitions.V8SerializationTag:
    """Returns the next non-padding serialization tag.

    Raises:
      errors.ParserError: when an invalid v8 tag is read.
    """
    while True:
      _, tag_value = self.decoder.ReadBytes(1)
      try:
        tag = definitions.V8SerializationTag(tag_value[0])
      except ValueError as error:
        raise errors.ParserError(
            f"Invalid v8 tag value {tag_value!r}"
        ) from error
      if tag != definitions.V8SerializationTag.PADDING:
        return tag

  def _ConsumeTag(self, peeked_tag: definitions.V8SerializationTag) -> None:
    """Consumes the next serialization tag.

    Args:
      peeked_tag: the tag that is expected to be read from the current position.

    Raises:
      errors.ParserError: if the peeked tag is not read.
    """
    tag = self._ReadTag()
    if tag != peeked_tag:
      raise errors.ParserError(f"Unexpected tag {tag} found.")

  def _ReadObject(self) -> Any:
    """Reads a Javascript object from the current position."""
    _, result = self._ReadObjectInternal()
    tag = self._PeekTag()
    if tag and tag == definitions.V8SerializationTag.ARRAY_BUFFER_VIEW:
      self._ConsumeTag(tag)
      result = self._ReadJSArrayBufferView(result)
      self.objects[self._GetNextId()] = result
    return result

  def _ReadObjectInternal(self) -> Tuple[definitions.V8SerializationTag, Any]:
    """Reads a Javascript object from the current position.

    Returns:
      a tuple of the serialization tag and the parsed object.
    """
    tag = self._ReadTag()
    parsed_object: Any = None

    if tag == definitions.V8SerializationTag.VERIFY_OBJECT_COUNT:
      _ = self.decoder.DecodeUint32Varint()
      parsed_object = self._ReadObject()
    elif tag == definitions.V8SerializationTag.UNDEFINED:
      parsed_object = types.Undefined()
    elif tag == definitions.V8SerializationTag.NULL:
      parsed_object = types.Null()
    elif tag == definitions.V8SerializationTag.TRUE:
      parsed_object = True
    elif tag == definitions.V8SerializationTag.FALSE:
      parsed_object = False
    elif tag == definitions.V8SerializationTag.INT32:
      _, parsed_object = self.decoder.DecodeInt32Varint()
    elif tag == definitions.V8SerializationTag.UINT32:
      _, parsed_object = self.decoder.DecodeUint32Varint()
    elif tag == definitions.V8SerializationTag.DOUBLE:
      _, parsed_object = self.decoder.DecodeDouble()
    elif tag == definitions.V8SerializationTag.BIGINT:
      parsed_object = self.ReadBigInt()
    elif tag == definitions.V8SerializationTag.UTF8_STRING:
      parsed_object = self.ReadUTF8String()
    elif tag == definitions.V8SerializationTag.ONE_BYTE_STRING:
      parsed_object = self.ReadOneByteString()
    elif tag == definitions.V8SerializationTag.TWO_BYTE_STRING:
      parsed_object = self.ReadTwoByteString()
    elif tag == definitions.V8SerializationTag.OBJECT_REFERENCE:
      _, object_id = self.decoder.DecodeUint32Varint()
      parsed_object = self.objects[object_id]
    elif tag == definitions.V8SerializationTag.BEGIN_JS_OBJECT:
      parsed_object = self._ReadJSObject()
    elif tag == definitions.V8SerializationTag.BEGIN_SPARSE_JS_ARRAY:
      parsed_object = self.ReadSparseJSArray()
    elif tag == definitions.V8SerializationTag.BEGIN_DENSE_JS_ARRAY:
      parsed_object = self.ReadDenseJSArray()
    elif tag == definitions.V8SerializationTag.DATE:
      parsed_object = self._ReadJSDate()
    elif tag in (
        definitions.V8SerializationTag.TRUE_OBJECT,
        definitions.V8SerializationTag.FALSE_OBJECT,
        definitions.V8SerializationTag.NUMBER_OBJECT,
        definitions.V8SerializationTag.BIGINT_OBJECT,
        definitions.V8SerializationTag.STRING_OBJECT,
    ):
      parsed_object = self._ReadJSPrimitiveWrapper(tag)
    elif tag == definitions.V8SerializationTag.REGEXP:
      parsed_object = self._ReadJSRegExp()
    elif tag == definitions.V8SerializationTag.BEGIN_JS_MAP:
      parsed_object = self._ReadJSMap()
    elif tag == definitions.V8SerializationTag.BEGIN_JS_SET:
      parsed_object = self._ReadJSSet()
    elif tag == definitions.V8SerializationTag.ARRAY_BUFFER:
      parsed_object = self._ReadJSArrayBuffer(
          is_shared=False, is_resizable=False
      )
    elif tag == definitions.V8SerializationTag.RESIZABLE_ARRAY_BUFFER:
      parsed_object = self._ReadJSArrayBuffer(
          is_shared=False, is_resizable=True
      )
    elif tag == definitions.V8SerializationTag.SHARED_ARRAY_BUFFER:
      parsed_object = self._ReadJSArrayBuffer(
          is_shared=True, is_resizable=False
      )
    elif tag == definitions.V8SerializationTag.ERROR:
      self._ReadJSError()
    elif tag == definitions.V8SerializationTag.WASM_MODULE_TRANSFER:
      self._ReadWasmModuleTransfer()
    elif tag == definitions.V8SerializationTag.WASM_MEMORY_TRANSFER:
      self._ReadWasmMemory()
    elif tag == definitions.V8SerializationTag.HOST_OBJECT:
      parsed_object = self.ReadHostObject()
    elif (
        tag == definitions.V8SerializationTag.SHARED_OBJECT
        and self.version >= 15
    ):
      self.ReadSharedObject()
    elif self.version < 13:
      self.decoder.stream.seek(-1, os.SEEK_CUR)
      parsed_object = self.ReadHostObject()
    else:
      parsed_object = None
    return tag, parsed_object

  def ReadString(self) -> str:
    """Reads a string from the current position.

    Raises:
      errors.ParserError: if the parsed object is not a string.
    """
    if self.version < 12:
      return self.ReadUTF8String()

    str_obj = self._ReadObject()
    if not isinstance(str_obj, str):
      raise errors.ParserError("Not a string")
    return str_obj

  def ReadBigInt(self) -> int:
    """Reads a Javascript Bigint from the current position."""
    bit_field = self.decoder.DecodeUint32Varint()[1]
    byte_count = bit_field >> 1
    signed = bool(bit_field & 0x1)
    _, bigint = self.decoder.DecodeInt(byte_count=byte_count)
    return -bigint if signed else bigint

  def ReadUTF8String(self) -> str:
    """Reads a UTF-8 string from the current position."""
    count = self.decoder.DecodeUint32Varint()[1]
    buffer = self.decoder.ReadBytes(count=count)[1]
    return buffer.decode("utf-8")

  def ReadOneByteString(self) -> str:
    """Reads a one-byte string from the current position.

    The raw bytes are decoded using latin-1 encoding.
    """
    length = self.decoder.DecodeUint32Varint()[1]
    buffer = self.decoder.ReadBytes(count=length)[1]
    return buffer.decode("latin-1")

  def ReadTwoByteString(self) -> str:
    """Reads a UTF-16-LE string from the current position."""
    length = self.decoder.DecodeUint32Varint()[1]
    buffer = self.decoder.ReadBytes(count=length)[1]
    return buffer.decode("utf-16-le")

  def ReadExpectedString(self) -> Optional[str]:
    """Reads a string from the current position, None if there is no tag or
      invalid byte length.

    A string can be UTF-8/one-byte/two-byte.

    Raises:
      errors.ParserError: if there is an unexpected serialization tag."""
    tag = self._ReadTag()
    if not tag:
      return None
    _, byte_length = self.decoder.DecodeUint32Varint()
    if not byte_length:
      return None

    _, raw_bytes = self.decoder.ReadBytes(count=byte_length)

    if tag == definitions.V8SerializationTag.ONE_BYTE_STRING:
      return raw_bytes.decode("latin-1")
    if tag == definitions.V8SerializationTag.TWO_BYTE_STRING:
      return raw_bytes.decode("utf-16-le")
    if tag == definitions.V8SerializationTag.UTF8_STRING:
      return raw_bytes.decode("utf-8")
    raise errors.ParserError(f"Unexpected serialization tag {tag}.")

  def _ReadJSObject(self) -> Dict[int, Any]:
    """Reads a JSObject from the current position.

    Raises:
      errors.ParserError: if an unexpected number of properties have been
        read.
    """
    next_id = self._GetNextId()
    js_object: Dict[int, Any] = {}

    num_properties = self._ReadJSObjectProperties(
        js_object, definitions.V8SerializationTag.END_JS_OBJECT
    )
    _, expected_number_properties = self.decoder.DecodeUint32Varint()
    if expected_number_properties != num_properties:
      raise errors.ParserError("Unexpected number of properties")

    self.objects[next_id] = js_object
    return js_object

  def _ReadJSObjectProperties(
      self,
      js_object: Union[Dict[int, Any], types.JSArray],
      end_tag: definitions.V8SerializationTag,
  ) -> int:
    """Reads key-value properties and sets them to the given js_object.

    Args:
      js_object: the Javascript object to set the parsed properties to.
      end_tag: the end tag that signifies there are no more properties to read.

    Returns:
      the number of properties that were read.
    """
    num_properties = 0
    while self._PeekTag() != end_tag:
      key = self._ReadObject()
      value = self._ReadObject()
      if isinstance(js_object, dict):
        js_object[key] = value
      else:
        js_object.properties[key] = value
      num_properties += 1
    self._ConsumeTag(end_tag)
    return num_properties

  def _GetNextId(self) -> int:
    """Gets the next object ID."""
    next_id = self.next_id
    self.next_id += 1
    return next_id

  def ReadSparseJSArray(self) -> types.JSArray:
    """Reads a sparse encoded JSArray from the current position.

    Raises:
      errors.ParserError: if there is an unexpected property or array length.
    """
    next_id = self._GetNextId()

    js_array = types.JSArray()
    _, length = self.decoder.DecodeUint32Varint()
    for _ in range(length):
      js_array.values.append(types.Undefined())

    num_properties = self._ReadJSObjectProperties(
        js_array.properties, definitions.V8SerializationTag.END_SPARSE_JS_ARRAY
    )
    _, expected_num_properties = self.decoder.DecodeUint32Varint()
    _, expected_length = self.decoder.DecodeUint32Varint()

    if num_properties != expected_num_properties:
      raise errors.ParserError("Unexpected property length")
    if length != expected_length:
      raise errors.ParserError("Unexpected array length")
    self.objects[next_id] = js_array
    return js_array

  def ReadDenseJSArray(self) -> types.JSArray:
    """Reads a dense encoded JSArray from the current position.

    Raises:
      errors.ParserError: if there is an unexpected property or array length.
    """
    next_id = self._GetNextId()

    js_array = types.JSArray()
    _, length = self.decoder.DecodeUint32Varint()
    for _ in range(length):
      tag = self._PeekTag()
      if tag == definitions.V8SerializationTag.THE_HOLE:
        self._ConsumeTag(tag)
        continue
      array_object = self._ReadObject()

      if self.version < 11 and isinstance(array_object, types.Undefined):
        continue
      js_array.values.append(array_object)

    num_properties = self._ReadJSObjectProperties(
        js_array.properties, definitions.V8SerializationTag.END_DENSE_JS_ARRAY
    )
    _, expected_num_properties = self.decoder.DecodeUint32Varint()
    _, expected_length = self.decoder.DecodeUint32Varint()
    if num_properties != expected_num_properties:
      raise errors.ParserError("Unexpected property length")
    if length != expected_length:
      raise errors.ParserError("Unexpected array length")
    self.objects[next_id] = js_array
    return js_array

  def _ReadJSDate(self) -> datetime:
    """Reads a JSDate from the current position."""
    next_id = self._GetNextId()

    _, value = self.decoder.DecodeDouble()
    result = datetime.utcfromtimestamp(value / 1000.0)
    self.objects[next_id] = result
    return result

  def _ReadJSPrimitiveWrapper(
      self, tag: definitions.V8SerializationTag
  ) -> Union[bool, float, int, str]:
    """Reads a Javascript wrapped primitive.

    Args:
      tag: the serialization tag.

    Returns:
      the parsed value.

    Raises:
      errors.ParserError: if the given tag is not a Javascript wrapped
        primitive.
    """
    next_id = self._GetNextId()

    if tag == definitions.V8SerializationTag.TRUE_OBJECT:
      value: Union[bool, float, int, str] = True
    elif tag == definitions.V8SerializationTag.FALSE_OBJECT:
      value = False
    elif tag == definitions.V8SerializationTag.NUMBER_OBJECT:
      _, value = self.decoder.DecodeDouble()
    elif tag == definitions.V8SerializationTag.BIGINT_OBJECT:
      value = self.ReadBigInt()
    elif tag == definitions.V8SerializationTag.STRING_OBJECT:
      value = self.ReadString()
    else:
      raise errors.ParserError(f"Invalid tag {tag}")

    self.objects[next_id] = value
    return value

  def _ReadJSRegExp(self) -> types.RegExp:
    """Reads a Javascript regular expression from the current position."""
    next_id = self._GetNextId()
    pattern = self.ReadString()
    _, flags = self.decoder.DecodeUint32Varint()  # TODO: verify flags
    regexp = types.RegExp(pattern=pattern, flags=str(flags))
    self.objects[next_id] = regexp
    return regexp

  def _ReadJSMap(self) -> Dict[Any, Any]:
    """Reads a Javascript map (dictionary) from the current position.

    Raises:
      errors.ParserError: on an unexpected map length.
    """
    next_id = self._GetNextId()
    js_map = {}

    tag = self._PeekTag()
    while tag != definitions.V8SerializationTag.END_JS_MAP:
      key = self._ReadObject()
      value = self._ReadObject()
      js_map[key] = value
      tag = self._PeekTag()
    self._ConsumeTag(definitions.V8SerializationTag.END_JS_MAP)

    _, expected_length = self.decoder.DecodeUint32Varint()
    if len(js_map) * 2 != expected_length:
      raise errors.ParserError("unexpected length")

    self.objects[next_id] = js_map
    return js_map

  def _ReadJSSet(self) -> Set[Any]:
    """Reads a Javascript set from the current position.

    Raises:
      errors.ParserError: on an unexpected map length.
    """
    next_id = self._GetNextId()

    js_set = set()
    tag = self._PeekTag()
    while tag != definitions.V8SerializationTag.END_JS_SET:
      element = self._ReadObject()
      js_set.add(element)
      tag = self._PeekTag()
    self._ConsumeTag(definitions.V8SerializationTag.END_JS_SET)

    _, expected_length = self.decoder.DecodeUint32Varint()
    if len(js_set) != expected_length:
      raise ValueError("unexpected length")

    self.objects[next_id] = js_set
    return js_set

  def _ReadJSArrayBuffer(self, is_shared: bool, is_resizable: bool) -> bytes:
    """Reads a Javascript ArrayBuffer from the current position.

    Args:
      is_shared: True if the buffer is shared, False otherwise.
      is_resizable: True if the buffer is resizable, False otherwise.
    """
    next_id = self._GetNextId()
    array_buffer = b""

    if is_shared:
      raise NotImplementedError("Shared ArrayBuffer not supported yet")

    _, byte_length = self.decoder.DecodeUint32Varint()
    max_byte_length = byte_length
    if is_resizable:
      _, max_byte_length = self.decoder.DecodeUint32Varint()
      if byte_length > max_byte_length:
        self.objects[next_id] = array_buffer
        return array_buffer
    if byte_length:
      _, array_buffer = self.decoder.ReadBytes(byte_length)

    self.objects[next_id] = array_buffer
    return array_buffer

  def _ReadJSArrayBufferView(self, buffer: bytes) -> ArrayBufferView:
    """Reads a JSArrayBufferView from the current position."""
    _, tag = self.decoder.ReadBytes(1)
    _, byte_offset = self.decoder.DecodeUint32Varint()
    _, byte_length = self.decoder.DecodeUint32Varint()

    if self.version >= 14:  # or version_13_broken_data_mode
      _, flags = self.decoder.DecodeUint32Varint()
    else:
      flags = 0

    return ArrayBufferView(
        buffer=buffer,
        tag=definitions.V8ArrayBufferViewTag(tag[0]),
        offset=byte_offset,
        length=byte_length,
        flags=flags,
    )

  def _ReadJSError(self) -> None:
    """Reads a Javascript error from the current position."""
    raise NotImplementedError("ValueDeserializer.ReadJSError")

  def _ReadWasmModuleTransfer(self) -> None:
    """Reads a Wasm module transfer object from the current position."""
    raise NotImplementedError("ValueDeserializer.ReadWasmModuleTransfer")

  def _ReadWasmMemory(self) -> None:
    """Reads a Wasm memory object from the current position."""
    raise NotImplementedError("ValueDeserializer.ReadWasmMemory")

  def ReadSharedObject(self) -> None:
    """Reads a shared object from the current position."""
    # _, shared_object_id = self.decoder.DecodeUint32Varint()
    raise NotImplementedError("ValueDeserializer.ReadSharedObject")

  def ReadHostObject(self) -> Any:
    """Reads a Host object using the delegate object.

    Raises:
      errors.ParserError: when the delegate object is None.
    """
    next_id = self._GetNextId()
    if not self.delegate:
      raise errors.ParserError("No delegate to read host object.")
    host_object = self.delegate.ReadHostObject()
    self.objects[next_id] = host_object
    return host_object

  @classmethod
  def FromBytes(cls, data: bytes, delegate: Any) -> Any:
    """Returns a deserialized javascript object from the data.

    Args:
      data: the data to deserialize/parse.
      delegate: an object to delegate parsing additional Javascript objects.

    Returns:
      A python representation of the parsed javascript object.

    Raises:
      errors.ParserError: if there is an invalid V8 javascript header.
    """
    stream = io.BytesIO(data)
    deserializer = cls(stream, delegate)
    if not deserializer.ReadHeader():
      raise errors.ParserError("Invalid V8 header")
    return deserializer.ReadObjectWrapper()



# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Types for indexeddb."""
from __future__ import annotations

import dataclasses
from typing import Any, Dict, List, Set


@dataclasses.dataclass
class JSArray:
  """A parsed Javascript array.

  A JavaScript array behaves like a Python list but allows assigning arbitrary
  properties.  The array is stored in the attribute __array__.

  Attributes:
    values: the array values.
    properties: the array properties.
  """

  values: List[Any] = dataclasses.field(default_factory=list)
  properties: Dict[Any, Any] = dataclasses.field(default_factory=dict)


@dataclasses.dataclass
class JSSet:
  """A parsed JavaScript set.

  A JavaScript set behaves like a Python set but allows assigning arbitrary
  properties.  The array is stored in the attribute __set__.

  Attributes:
    values: the set values.
    properties: the set properties.
  """

  values: Set[Any] = dataclasses.field(default_factory=set)
  properties: Dict[Any, Any] = dataclasses.field(default_factory=dict)


@dataclasses.dataclass
class Null:
  """A parsed JavaScript Null."""


@dataclasses.dataclass
class RegExp:
  """A parsed JavaScript RegExp.

  Attributes:
    pattern: the pattern.
    flags: the flags.
  """

  pattern: str
  flags: str


@dataclasses.dataclass
class Undefined:
  """A JavaScript undef."""



# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Definitions for Firefox IndexedDB."""
from enum import IntEnum


class IndexedDBKeyType(IntEnum):
  """IndexedDB Key Types."""

  TERMINATOR = 0
  FLOAT = 0x10
  DATE = 0x20
  STRING = 0x30
  BINARY = 0x40
  ARRAY = 0x50


MAX_ARRAY_COLLAPSE = 3
MAX_RECURSION_DEPTH = 64
MAX_LENGTH = (1 << 30) - 2
ONE_BYTE_LIMIT = 0x7E
TWO_BYTE_LIMIT = 0x3FFF + 0x7F
ONE_BYTE_ADJUST = 1
TWO_BYTE_ADJUST = -0x7F
THREE_BYTE_SHIFT = 6


class StructuredDataType(IntEnum):
  """Structured Data Types."""

  FLOAT_MAX = 0xFFF00000
  HEADER = 0xFFF10000
  NULL = 0xFFFF0000
  UNDEFINED = 0xFFFF0001
  BOOLEAN = 0xFFFF0002
  INT32 = 0xFFFF0003
  STRING = 0xFFFF0004
  DATE_OBJECT = 0xFFFF0005
  REGEXP_OBJECT = 0xFFFF0006
  ARRAY_OBJECT = 0xFFFF0007
  OBJECT_OBJECT = 0xFFFF0008
  ARRAY_BUFFER_OBJECT_V2 = 0xFFFF0009
  BOOLEAN_OBJECT = 0xFFFF000A
  STRING_OBJECT = 0xFFFF000B
  NUMBER_OBJECT = 0xFFFF000C
  BACK_REFERENCE_OBJECT = 0xFFFF000D
  DO_NOT_USE_1 = 0xFFFF000E
  DO_NOT_USE_2 = 0xFFFF000F
  TYPED_ARRAY_OBJECT_V2 = 0xFFFF0010
  MAP_OBJECT = 0xFFFF0011
  SET_OBJECT = 0xFFFF0012
  END_OF_KEYS = 0xFFFF0013
  DO_NOT_USE_3 = 0xFFFF0014
  DATA_VIEW_OBJECT_V2 = 0xFFFF0015
  SAVED_FRAME_OBJECT = 0xFFFF0016
  JSPRINCIPALS = 0xFFFF0017
  NULL_JSPRINCIPALS = 0xFFFF0018
  RECONSTRUCTED_SAVED_FRAME_PRINCIPALS_IS_SYSTEM = 0xFFFF0019
  RECONSTRUCTED_SAVED_FRAME_PRINCIPALS_IS_NOT_SYSTEM = 0xFFFF001A
  SHARED_ARRAY_BUFFER_OBJECT = 0xFFFF001B
  SHARED_WASM_MEMORY_OBJECT = 0xFFFF001C
  BIGINT = 0xFFFF001D
  BIGINT_OBJECT = 0xFFFF001E
  ARRAY_BUFFER_OBJECT = 0xFFFF001F
  TYPED_ARRAY_OBJECT = 0xFFFF0020
  DATA_VIEW_OBJECT = 0xFFFF0021
  ERROR_OBJECT = 0xFFFF0022
  RESIZABLE_ARRAY_BUFFER_OBJECT = 0xFFFF0023
  GROWABLE_SHARED_ARRAY_BUFFER_OBJECT = 0xFFFF0024
  TYPED_ARRAY_V1_INT8 = 0xFFFF0100
  TYPED_ARRAY_V1_UINT8 = 0xFFFF0101
  TYPED_ARRAY_V1_INT16 = 0xFFFF0102
  TYPED_ARRAY_V1_UINT16 = 0xFFFF0103
  TYPED_ARRAY_V1_INT32 = 0xFFFF0104
  TYPED_ARRAY_V1_UINT32 = 0xFFFF0105
  TYPED_ARRAY_V1_FLOAT32 = 0xFFFF0106
  TYPED_ARRAY_V1_FLOAT64 = 0xFFFF0107
  TYPED_ARRAY_V1_UINT8_CLAMPED = 0xFFFF0108
  TRANSFER_MAP_HEADER = 0xFFFF0200
  TRANSFER_MAP_PENDING_ENTRY = 0xFFFF0201
  TRANSFER_MAP_ARRAY_BUFFER = 0xFFFF0202
  TRANSFER_MAP_STORED_ARRAY_BUFFER = 0xFFFF0203
  TRANSFER_MAP_END_OF_BUILTIN_TYPES = 0xFFFF0204


class StructuredCloneTags(IntEnum):
  """Structured Clone Tags."""

  BLOB = 0xFFFF8001
  FILE_WITHOUT_LASTMODIFIEDDATE = 0xFFFF8002
  FILELIST = 0xFFFF8003
  MUTABLEFILE = 0xFFFF8004
  FILE = 0xFFFF8005
  WASM_MODULE = 0xFFFF8006
  IMAGEDATA = 0xFFFF8007
  DOMPOINT = 0xFFFF8008
  DOMPOINTREADONLY = 0xFFFF8009
  CRYPTOKEY = 0xFFFF800A
  NULL_PRINCIPAL = 0xFFFF800B
  SYSTEM_PRINCIPAL = 0xFFFF800C
  CONTENT_PRINCIPAL = 0xFFFF800D
  DOMQUAD = 0xFFFF800E
  RTCCERTIFICATE = 0xFFFF800F
  DOMRECT = 0xFFFF8010
  DOMRECTREADONLY = 0xFFFF8011
  EXPANDED_PRINCIPAL = 0xFFFF8012
  DOMMATRIX = 0xFFFF8013
  URLSEARCHPARAMS = 0xFFFF8014
  DOMMATRIXREADONLY = 0xFFFF8015
  DOMEXCEPTION = 0xFFFF80016
  EMPTY_SLOT_9 = 0xFFFF8017
  STRUCTUREDCLONETESTER = 0xFFFF8018
  FILESYSTEMHANDLE = 0xFFFF8019
  FILESYSTEMFILEHANDLE = 0xFFFF801A
  FILESYSTEMDIRECTORYHANDLE = 0xFFFF801B
  IMAGEBITMAP = 0xFFFF801C
  MAP_MESSAGEPORT = 0xFFFF801D
  FORMDATA = 0xFFFF801E
  CANVAS = 0xFFFF801F  # This tag is for OffscreenCanvas.
  DIRECTORY = 0xFFFF8020
  INPUTSTREAM = 0xFFFF8021
  STRUCTURED_CLONE_HOLDER = 0xFFFF8022
  BROWSING_CONTEXT = 0xFFFF8023
  CLONED_ERROR_OBJECT = 0xFFFF8024
  READABLESTREAM = 0xFFFF8025
  WRITABLESTREAM = 0xFFFF8026
  TRANSFORMSTREAM = 0xFFFF8027
  VIDEOFRAME = 0xFFFF8028
  ENCODEDVIDEOCHUNK = 0xFFFF8029
  AUDIODATA = 0xFFFF8030
  ENCODEDAUDIOCHUNK = 0xFFFF8031


FRAME_HEADER = b"\xff\x06\x00\x00sNaPpY"



# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Parsers for Gecko encoded JavaScript values."""
from __future__ import annotations

import dataclasses
import datetime
import io
import struct
from typing import Any, List, Tuple, Union

import snappy

from dfindexeddb import errors, utils
from dfindexeddb.indexeddb import types
from dfindexeddb.indexeddb.firefox import definitions


@dataclasses.dataclass
class IDBKey(utils.FromDecoderMixin):
  """An IndexedDB Key.

  Attributes:
    offset: the offset of the key.
    type: the type of the key.
    value: the value of the key.
  """

  offset: int
  type: definitions.IndexedDBKeyType
  value: Union[bytes, datetime.datetime, float, list[Any], str, None]

  @classmethod
  def FromDecoder(
      cls, decoder: utils.StreamDecoder, base_offset: int = 0
  ) -> IDBKey:
    """Decodes an IndexedDB key from the current position of the decoder.

    Args:
      decoder: the decoder.
      base_offset: the base offset.

    Returns:
      The IndexedDB Key.
    """
    offset, key_type, data = KeyDecoder(decoder.stream).DecodeKey()
    return cls(offset=offset + base_offset, type=key_type, value=data)


class KeyDecoder(utils.StreamDecoder):
  """A helper class to decode the Gecko-encoded key value."""

  def DecodeKey(self) -> Tuple[int, definitions.IndexedDBKeyType, Any]:
    """Decodes the key.

    Returns:
      a tuple comprising the offset where the key was decoded, the key type and
          the key value.
    """
    offset, key_type_bytes = self.PeekBytes(1)
    value = self._DecodeJSValInternal(key_type_bytes[0])
    if key_type_bytes[0] >= definitions.IndexedDBKeyType.ARRAY:
      key_type = definitions.IndexedDBKeyType.ARRAY
    else:
      key_type = definitions.IndexedDBKeyType(key_type_bytes[0])
    return offset, key_type, value

  def _DecodeJSValInternal(
      self, key_type: int, type_offset: int = 0, recursion_depth: int = 0
  ) -> Any:
    """Decodes a Javascript Value.

    Args:
      key_type: the key type byte value.
      type_offset: the type_offset.
      recursion_depth: the current recursion depth.

    Returns:
      the value.

    Raises:
      ParserError: when reaching maximum recursion depth or when an known key
          type has been parsed.
    """
    if recursion_depth == definitions.MAX_RECURSION_DEPTH:
      raise errors.ParserError("Reached Maximum Recursion Depth")

    value: Any = None
    if key_type - type_offset >= definitions.IndexedDBKeyType.ARRAY:
      type_offset += definitions.IndexedDBKeyType.ARRAY
      if (
          type_offset
          == definitions.IndexedDBKeyType.ARRAY * definitions.MAX_ARRAY_COLLAPSE
      ):
        _ = self.ReadBytes(1)
        type_offset = 0
      value = []
      while self.NumRemainingBytes() and (key_type - type_offset) != (
          definitions.IndexedDBKeyType.TERMINATOR
      ):
        value.append(
            self._DecodeJSValInternal(
                key_type, type_offset, recursion_depth + 1
            )
        )
        type_offset = 0
        if self.NumRemainingBytes():
          _, key_bytes = self.PeekBytes(1)
          key_type = key_bytes[0]

      if self.NumRemainingBytes():
        _ = self.ReadBytes(1)  # consume terminator byte

    elif key_type - type_offset == definitions.IndexedDBKeyType.STRING:
      _, value = self._DecodeAsStringy()
      value = value.decode("utf-8")  # TODO: test other text encoding types
    elif key_type - type_offset == definitions.IndexedDBKeyType.DATE:
      _, value = self._DecodeDate()
    elif key_type - type_offset == definitions.IndexedDBKeyType.FLOAT:
      _, value = self._DecodeFloat()
    elif key_type - type_offset == definitions.IndexedDBKeyType.BINARY:
      _, value = self._DecodeAsStringy()
    else:
      raise errors.ParserError(
          f"Unknown key type parsed: {key_type - type_offset}"
      )
    return value

  def _ReadUntilNull(self) -> bytearray:
    """Read bytes until a null (terminator) byte is encountered."""
    result = bytearray()
    num_remaining_bytes = self.NumRemainingBytes()
    while num_remaining_bytes:
      _, c = self.ReadBytes(1)
      if c == b"\x00":
        break
      result += c
      num_remaining_bytes -= 1
    return result

  def _DecodeAsStringy(self, element_size: int = 1) -> Tuple[int, bytearray]:
    """Decodes a string buffer.

    Args:
      element_size: parse string as a multi-byte string.

    Returns:
      A tuple of the offset and the decoded string.

    Raises:
      ParserError: if an invalid type is read or an unsupported
          encoded byte is encountered
    """
    offset, type_int = self.DecodeUint8()
    if type_int % definitions.IndexedDBKeyType.ARRAY.value not in (
        definitions.IndexedDBKeyType.STRING.value,
        definitions.IndexedDBKeyType.BINARY.value,
    ):
      raise errors.ParserError(
          "Invalid type", hex(type_int), hex(type_int % 0x50)
      )

    i = 0
    result = self._ReadUntilNull()
    while i < len(result):
      if not result[i] & 0x80:
        result[i] -= definitions.ONE_BYTE_ADJUST
      elif element_size == 2 or not result[i] & 0x40:
        c = struct.unpack_from(">H", result, i)[0]
        d = c - 0x8000 - definitions.TWO_BYTE_ADJUST
        struct.pack_into(">H", result, i, d)
      else:
        raise errors.ParserError("Unsupported byte")  # TODO: add support.
      i += 1
    return offset + 1, result

  def _DecodeFloat(self) -> Tuple[int, float]:
    """Decodes a float.

    Returns:
      A tuple of the offset and a (double-precision) float value.

    Raises:
      ParserError if an invalid type is read
    """
    _, type_byte = self.DecodeUint8()
    if type_byte % definitions.IndexedDBKeyType.ARRAY not in (
        definitions.IndexedDBKeyType.FLOAT,
        definitions.IndexedDBKeyType.DATE,
    ):
      raise errors.ParserError(
          "Invalid type", hex(type_byte), hex(type_byte % 0x50)
      )

    float_offset, number_bytes = self.ReadBytes(
        min(8, self.NumRemainingBytes())
    )

    # pad to 64-bits
    if len(number_bytes) != 8:
      number_bytes += b"\x00" * (8 - len(number_bytes))

    int_value = struct.unpack(">q", number_bytes)[0]

    if int_value & 0x8000000000000000:
      int_value = int_value & 0x7FFFFFFFFFFFFFFF
    else:
      int_value = -int_value
    return float_offset, struct.unpack(">d", struct.pack(">q", int_value))[0]

  def _DecodeBinary(self) -> Tuple[int, bytes]:
    """Decodes a binary value.

    Returns:
      A tuple of the offset and the bytearray.

    Raises:
      ParserError: when an invalid type byte is encountered.
    """
    _, type_byte = self.ReadBytes(1)
    if (
        type_byte[0] % definitions.IndexedDBKeyType.ARRAY
        != definitions.IndexedDBKeyType.BINARY
    ):
      raise errors.ParserError(
          f"Invalid type {type_byte[0] % definitions.IndexedDBKeyType.ARRAY}"
      )
    values = bytearray()
    offset = self.stream.tell()
    while self.NumRemainingBytes():
      _, value = self.ReadBytes(1)
      if value[0] == definitions.IndexedDBKeyType.TERMINATOR:
        break
      values.append(value[0])
    return offset, bytes(values)

  def _DecodeDate(self) -> Tuple[int, datetime.datetime]:
    """Decodes a date.

    Returns:
      A tuple of the offset and the decoded datetime.
    """
    offset, value = self._DecodeFloat()
    return offset, datetime.datetime.fromtimestamp(
        value / 1000, tz=datetime.timezone.utc
    )


class JSStructuredCloneDecoder(utils.FromDecoderMixin):
  """Decodes a serialized JavaScript value.

  Attributes:
    decoder: for decoding the buffer.
    all_objects: all the decoded objects.
  """

  def __init__(self, decoder: utils.StreamDecoder):
    """Initializes the reader.

    Args:
      buffer: the serialized bytes.
    """
    self.decoder = decoder
    self.all_objects: List[Any] = []
    self._object_stack: List[Any] = []

  def _PeekTagAndData(self) -> tuple[int, int]:
    """Peeks Tag and Data values."""
    _, pair_bytes = self.decoder.PeekBytes(8)
    pair = int.from_bytes(pair_bytes, byteorder="little", signed=False)
    tag = pair >> 32
    data = pair & 0x00000000FFFFFFFF
    return tag, data

  def _DecodeTagAndData(self) -> tuple[int, int]:
    """Decodes Tag and Data values."""
    _, pair = self.decoder.DecodeUint64()
    tag = pair >> 32
    data = pair & 0x00000000FFFFFFFF
    return tag, data

  def _DecodeString(self, data: int) -> str:
    """Decodes a String.

    Args:
      data: the data value.

    Returns:
      the decoded string.

    Raises:
      ParserError: if the serialized string length is too large.
    """
    number_chars = data & 0x7FFFFFFF
    latin1 = data & 0x80000000
    if number_chars > definitions.MAX_LENGTH:
      raise errors.ParserError("Bad serialized string length")

    if latin1:
      _, char_bytes = self.decoder.ReadBytes(number_chars)
      chars = char_bytes.decode("latin-1")
    else:
      _, char_bytes = self.decoder.ReadBytes(number_chars * 2)
      chars = char_bytes.decode("utf-16-le")

    return chars

  def _DecodeBigInt(self, data: int) -> int:
    """Decodes a BigInteger.

    Args:
      data: the data value

    Returns:
      the decoded integer.
    """
    length = data & 0x7FFFFFFF
    is_negative = data & 0x80000000
    if length == 0:
      return 0
    contents = bytearray()
    for _ in range(length):
      _, element = self.decoder.ReadBytes(8)
      contents.extend(element)
    return int.from_bytes(
        contents, byteorder="little", signed=bool(is_negative)
    )

  def _DecodeTypedArray(
      self, array_type: int, number_elements: int, is_version1: bool = False
  ) -> Any:
    """Decodes a typed array.

    Args:
      array_type: the array type.
      number_elements: the number of elements.
      is_version1: True if the typed array is of version 1 type.

    Returns:
      The typed array.
    """
    self.all_objects.append(None)
    dummy_index = len(self.all_objects) - 1

    is_auto_length = number_elements == -1
    if is_auto_length:
      number_elements = 0

    if is_version1:
      self._DecodeV1ArrayBuffer(array_type, number_elements)
      byte_offset = 0  # pylint: disable=unused-variable
    else:
      value = self._StartRead()
      _, byte_offset = self.decoder.DecodeUint64()  # pylint: disable=unused-variable

    self.all_objects[dummy_index] = value  # pylint: disable=possibly-used-before-assignment
    return value

  def _DecodeArrayBuffer(self, buffer_type: int, data: int) -> bytes:
    """Decodes an ArrayBuffer.

    Args:
      buffer_type: the data type.
      data: the data value.

    Returns:
      the array buffer.
    """
    if buffer_type == definitions.StructuredDataType.ARRAY_BUFFER_OBJECT:
      _, number_bytes = self.decoder.DecodeUint64()
    elif (
        buffer_type
        == definitions.StructuredDataType.RESIZABLE_ARRAY_BUFFER_OBJECT
    ):
      _, number_bytes = self.decoder.DecodeUint64()
      _, max_bytes = self.decoder.DecodeUint64()  # pylint: disable=unused-variable
    else:
      number_bytes = data

    _, buffer = self.decoder.ReadBytes(number_bytes)
    return buffer

  def _DecodeV1ArrayBuffer(self, array_type: int, number_elements: int) -> None:
    """Decodes a V1 ArrayBuffer.

    Not currently implemented.

    Raises:
      NotImplementedError: because it's not yet supported.
    """
    raise NotImplementedError("V1 Array Buffer not supported")

  def DecodeValue(self) -> Any:
    """Decodes a Javascript value."""
    if not self._DecodeHeader():
      return None

    tag, _ = self._PeekTagAndData()
    if tag == definitions.StructuredDataType.TRANSFER_MAP_HEADER:
      raise errors.ParserError("Transfer Map is not supported.")

    value = self._StartRead()

    while self._object_stack:
      offset = self.decoder.stream.tell() % 8
      if offset:
        self.decoder.stream.seek(8 - offset, io.SEEK_CUR)
      tag, _ = self._PeekTagAndData()
      obj = self._object_stack[-1]
      if tag == definitions.StructuredDataType.END_OF_KEYS:
        self._object_stack.pop()
        _, _ = self._DecodeTagAndData()
        continue

      expect_key_value_pairs = isinstance(obj, (dict, set))
      key = self._StartRead()
      if isinstance(key, types.Null) and expect_key_value_pairs:
        self._object_stack.pop()
        continue

      if isinstance(obj, types.JSSet):
        obj.values.add(key)
      elif isinstance(obj, types.JSArray):
        field = self._StartRead()
        obj.values[key] = field
      else:
        field = self._StartRead()
        obj[key] = field

    return value

  def _DecodeHeader(self) -> bool:
    """Decodes the header.

    Returns:
      True if the correct header tag is read.

    Raises:
      ParserError: when a header tag is not found.
    """
    tag, _ = self._DecodeTagAndData()
    if tag != definitions.StructuredDataType.HEADER:
      raise errors.ParserError("Header tag not found.")
    return True

  def _DecodeRegexp(self, flags: int) -> types.RegExp:
    """Decodes a regular expression.

    Args:
      flags: the regular expression flags.

    Raises:
      ParserError: when the string tag is not found.
    """
    tag, string_data = self._DecodeTagAndData()
    if tag != definitions.StructuredDataType.STRING:
      raise errors.ParserError("String tag not found")

    pattern = self._DecodeString(string_data)
    return types.RegExp(pattern=pattern, flags=str(flags))

  def _StartRead(self) -> Any:
    """Reads the start of a serialized value.

    Raises:
      ParserError: when an unsupported StructuredDataType,
          StructuredCloneTag or unknown tag is read.
    """
    value: Any = None
    tag, data = self._DecodeTagAndData()
    if tag == definitions.StructuredDataType.NULL:
      value = types.Null()
    elif tag == definitions.StructuredDataType.UNDEFINED:
      value = types.Undefined()
    elif tag == definitions.StructuredDataType.INT32:
      value = data
    elif tag == definitions.StructuredDataType.BOOLEAN:
      value = bool(data)
    elif tag == definitions.StructuredDataType.BOOLEAN_OBJECT:
      value = bool(data)
      self.all_objects.append(value)
    elif tag == definitions.StructuredDataType.STRING:
      value = self._DecodeString(data)
    elif tag == definitions.StructuredDataType.STRING_OBJECT:
      value = self._DecodeString(data)
      self.all_objects.append(value)
    elif tag == definitions.StructuredDataType.NUMBER_OBJECT:
      _, value = self.decoder.DecodeDouble()
      self.all_objects.append(value)
    elif tag == definitions.StructuredDataType.BIGINT:
      value = self._DecodeBigInt(data)
    elif tag == definitions.StructuredDataType.BIGINT_OBJECT:
      value = self._DecodeBigInt(data)
      self.all_objects.append(value)
    elif tag == definitions.StructuredDataType.DATE_OBJECT:
      _, timestamp = self.decoder.DecodeDouble()
      value = datetime.datetime.fromtimestamp(
          timestamp / 1000, tz=datetime.timezone.utc
      )
      self.all_objects.append(value)
    elif tag == definitions.StructuredDataType.REGEXP_OBJECT:
      value = self._DecodeRegexp(data)
      self.all_objects.append(value)
    elif tag == definitions.StructuredDataType.ARRAY_OBJECT:
      value = types.JSArray()
      for _ in range(data):
        value.values.append(None)
      self._object_stack.append(value)
      self.all_objects.append(value)
    elif tag == definitions.StructuredDataType.OBJECT_OBJECT:
      value = {}
      self._object_stack.append(value)
      self.all_objects.append(value)
    elif tag == definitions.StructuredDataType.BACK_REFERENCE_OBJECT:
      value = self.all_objects[data]
    elif tag == definitions.StructuredDataType.ARRAY_BUFFER_OBJECT_V2:
      value = self._DecodeArrayBuffer(tag, data)
      self.all_objects.append(value)
    elif tag == definitions.StructuredDataType.ARRAY_BUFFER_OBJECT:
      value = self._DecodeArrayBuffer(tag, data)
      self.all_objects.append(value)
    elif tag == definitions.StructuredDataType.RESIZABLE_ARRAY_BUFFER_OBJECT:
      value = self._DecodeArrayBuffer(tag, data)
      self.all_objects.append(value)
    elif tag == definitions.StructuredDataType.TYPED_ARRAY_OBJECT_V2:
      _, array_type = self.decoder.DecodeUint64()
      self._DecodeTypedArray(array_type, data)
    elif tag == definitions.StructuredDataType.TYPED_ARRAY_OBJECT:
      _, number_elements = self.decoder.DecodeUint64()
      value = self._DecodeTypedArray(data, number_elements)
    elif tag == definitions.StructuredDataType.MAP_OBJECT:
      value = {}
      self._object_stack.append(value)
      self.all_objects.append(value)
    elif tag == definitions.StructuredDataType.SET_OBJECT:
      value = types.JSSet()
      self._object_stack.append(value)
      self.all_objects.append(value)
    elif tag <= definitions.StructuredDataType.FLOAT_MAX:
      value = struct.unpack("<d", struct.pack("<II", data, tag))[0]
    elif (
        definitions.StructuredDataType.TYPED_ARRAY_V1_INT8
        <= tag
        <= definitions.StructuredDataType.TYPED_ARRAY_V1_UINT8_CLAMPED
    ):
      value = self._DecodeTypedArray(tag, data)
    elif tag in definitions.StructuredDataType.__members__.values():
      raise errors.ParserError(
          "Unsupported StructuredDataType", definitions.StructuredDataType(tag)
      )
    elif tag in definitions.StructuredCloneTags.__members__.values():
      raise errors.ParserError(
          "Unsupported StructuredCloneTag", definitions.StructuredCloneTags(tag)
      )
    else:
      raise errors.ParserError("Unknown tag", hex(tag))

    # align the stream to an 8 byte boundary
    offset = self.decoder.stream.tell() % 8
    if offset:
      _, slack_bytes = self.decoder.ReadBytes(8 - offset)  # pylint: disable=unused-variable

    return value

  @classmethod
  def FromDecoder(
      cls, decoder: utils.StreamDecoder, base_offset: int = 0
  ) -> Any:
    """Decodes the parsed JavaScript object from the StreamDecoder.

    Args:
      decoder: the StreamDecoder.
      base_offset: the base offset.

    Returns:
      The class instance.
    """
    return cls(decoder).DecodeValue()

  @classmethod
  def FromBytes(cls, raw_data: bytes, base_offset: int = 0) -> Any:
    """Returns the parsed JavaScript object from the raw bytes.

    Args:
      raw_data: the raw data to deserialize/parse.

    Returns:
      A python representation of the parsed JavaScript object.
    """
    if raw_data.startswith(definitions.FRAME_HEADER):
      uncompressed_data = bytearray()
      pos = len(definitions.FRAME_HEADER)
      while pos < len(raw_data):
        is_uncompressed = raw_data[pos]
        block_size = int.from_bytes(
            raw_data[pos + 1 : pos + 4], byteorder="little", signed=False
        )
        masked_checksum = int.from_bytes(  # pylint: disable=unused-variable
            raw_data[pos + 4 : pos + 9], byteorder="little", signed=False
        )
        if is_uncompressed:
          uncompressed_data += raw_data[pos + 8 : pos + 8 + block_size - 4]
        else:
          uncompressed_data += snappy.decompress(
              raw_data[pos + 8 : pos + 8 + block_size - 4]
          )
        pos += block_size + 4
    else:
      uncompressed_data = snappy.decompress(raw_data)
    stream = io.BytesIO(uncompressed_data)

    return cls.FromStream(stream=stream, base_offset=base_offset)



# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Parsers for blink javascript serialized objects."""
from __future__ import annotations

import io
from dataclasses import dataclass
from typing import Any, Optional, Union

import snappy

from dfindexeddb import errors, utils
from dfindexeddb.indexeddb.chromium import definitions, v8

_MS_PER_SECOND = 1000


@dataclass
class AudioData:
  """A Javascript AudioData class.

  Attributes:
    audio_frame_index: the Audio Frame index.
  """

  audio_frame_index: int


@dataclass
class BaseIndex:
  """A base index class."""

  index: int


@dataclass
class BitMap:
  """A Javascript BitMap."""


@dataclass
class Blob:
  """A Javascript Blob class type.

  Attributes:
    uuid: the UUID of the Blob.
    type: the type of the Blob.
    size: the size of the Blob.
  """

  uuid: str
  type: str
  size: int


@dataclass
class BlobIndex(BaseIndex):
  """A Javascript BlobIndex class type."""


@dataclass
class CryptoKey:
  """A Javascript CryptoKey class type.

  Attributes:
    algorithm: the CryptoKey algorithm.
    key_type: the CryptoKey type.
    extractable: True if the CryptoKey is extractable.
    usages: the CryptoKey usage.
    key_data: the raw key data.
  """

  algorithm_parameters: dict[str, Any]
  key_type: Union[
      definitions.WebCryptoKeyType, definitions.AsymmetricCryptoKeyType
  ]
  extractable: bool
  usages: definitions.CryptoKeyUsage
  key_data: bytes


@dataclass
class DOMException:
  """A Javascript DOMException.

  Attributes:
    name: the name.
    message: the message.
    stack_unused: the stack unused.
  """

  name: str
  message: str
  stack_unused: str


@dataclass
class DOMFileSystem:
  """A Javascript DOMFileSystem.

  Attributes:
    raw_type: the raw type.
    name: the name.
    root_url: the root URL.
  """

  raw_type: int
  name: str
  root_url: str


@dataclass
class DOMMatrix2D:
  """A Javascript DOMMatrix2D.

  Attributes:
    values: the values.
  """

  values: list[float]


@dataclass
class DOMMatrix2DReadOnly(DOMMatrix2D):
  """A Javascript Read-Only DOMMatrix2D."""


@dataclass
class DOMMatrix:
  """A Javascript DOMMatrix.

  Attributes:
    values: the values.
  """

  values: list[float]


@dataclass
class DOMMatrixReadOnly(DOMMatrix):
  """A Javascript Read-Only DOMMatrix."""


@dataclass
class DOMPoint:
  """A Javascript DOMPoint.

  Attributes:
    x: the X coordinate.
    y: the Y coordinate.
    z: the Z coordinate.
    w: the W coordinate.
  """

  x: float
  y: float
  z: float
  w: float


@dataclass
class DOMPointReadOnly(DOMPoint):
  """A Javascript Read-Only DOMPoint."""


@dataclass
class DOMQuad:
  """A Javascript DOMQuad.

  Attributes:
    p1: the first point.
    p2: the second point.
    p3: the third point.
    p4: the fourth point.
  """

  p1: DOMPoint
  p2: DOMPoint
  p3: DOMPoint
  p4: DOMPoint


@dataclass
class DOMRect:
  """A Javascript DOMRect.

  Attributes:
    x: the X coordinate.
    y: the Y coordinate.
    width: the width.
    height: the height.
  """

  x: float
  y: float
  width: float
  height: float


@dataclass
class DOMRectReadOnly(DOMRect):
  """A Javascript Read-only DOMRect."""


@dataclass
class EncodedAudioChunk(BaseIndex):
  """A Javascript EncodedAudioChunk."""


@dataclass
class EncodedVideoChunk(BaseIndex):
  """A Javascript EncodedVideoChunk."""


@dataclass
class File:
  """A Javascript File.

  Attributes:
    path: the file path.
    name: the file name.
    relative_path: the file relative path.
    uuid: the file UUID.
    type: the file type.
    has_snapshot: True if the file has a snapshot.
    size: the file size.
    last_modified_ms: the file last modified in milliseconds.
    is_user_visible: True if the file is user visible.
  """

  path: str
  name: Optional[str]
  relative_path: Optional[str]
  uuid: str
  type: str
  has_snapshot: int
  size: Optional[int]
  last_modified_ms: Optional[float]
  is_user_visible: int


@dataclass
class FileIndex(BaseIndex):
  """A Javascript FileIndex."""


@dataclass
class FileList:
  """A Javascript FileList.

  Attributes:
    files: the list of files.
  """

  files: list[File]


@dataclass
class FileListIndex:
  """A Javascript FileListIndex.

  Attributes:
    file_indexes: the list of file indexes.
  """

  file_indexes: list[FileIndex]


@dataclass
class FileSystemHandle:
  """A Javascript FileSystemHandle.

  Attributes:
    name: the file system handle name.
    token_index: the file system token index.
  """

  name: str
  token_index: int


@dataclass
class ImageBitmapTransfer(BaseIndex):
  """A Javascript ImageBitmapTransfer."""


@dataclass
class MediaSourceHandle(BaseIndex):
  """A Javascript MediaSourceHandle."""


@dataclass
class MessagePort(BaseIndex):
  """A Javascript MessagePort."""


@dataclass
class MojoHandle(BaseIndex):
  """A Javascript MojoHandle."""


@dataclass
class ReadableStreamTransfer(BaseIndex):
  """A Javascript ReadableStreamTransfer."""


@dataclass
class RTCEncodedAudioFrame(BaseIndex):
  """A Javascript RTCEncodedAudioFrame."""


@dataclass
class RTCEncodedVideoFrame(BaseIndex):
  """A Javascript RTCEncodedVideoFrame."""


@dataclass
class WriteableStreamTransfer(BaseIndex):
  """A Javascript WriteableStreamTransfer."""


@dataclass
class TransformStreamTransfer(BaseIndex):
  """A Javascript TransformStreamTransfer."""


@dataclass
class OffscreenCanvasTransfer:
  """A Javascript Offscreen Canvas Transfer."""

  width: int
  height: int
  canvas_id: int
  client_id: int
  sink_id: int
  filter_quality: int


@dataclass
class UnguessableToken:
  """A Javascript Unguessable Token.

  Attributes:
    high: the high part.
    low: the low part.
  """

  high: int
  low: int


@dataclass
class VideoFrame(BaseIndex):
  """A Javascript VideoFrame."""


class V8ScriptValueDecoder:
  """A Blink V8 Serialized Script Value (SSV) decoder.

  Attributes:
    deserializer (v8.ValueDeserializer): the V8 value deserializer.
    raw_data (bytes): the raw bytes containing the serialized javascript
        value.
    version (int): the blink version.
  """

  _MIN_VERSION_FOR_SEPARATE_ENVELOPE = 16

  # As defined in trailer_reader.h
  _MIN_WIRE_FORMAT_VERSION = 21

  def __init__(self, raw_data: bytes):
    self.raw_data = raw_data
    self.deserializer: v8.ValueDeserializer = v8.ValueDeserializer(
        io.BytesIO(self.raw_data), delegate=self
    )
    self.version = 0
    self.trailer_offset: int | None = None
    self.trailer_size: int | None = None

  def _ReadVersionEnvelope(self) -> int:
    """Reads the Blink version envelope.

    Returns:
      The number of bytes consumed reading the Blink version envelope or zero
          if no blink envelope is detected.
    """
    if not self.raw_data:
      return 0

    decoder = utils.StreamDecoder(io.BytesIO(self.raw_data))
    _, tag = decoder.DecodeUint8()
    if tag != definitions.BlinkSerializationTag.VERSION:
      return 0

    _, self.version = decoder.DecodeUint32Varint()
    if self.version < self._MIN_VERSION_FOR_SEPARATE_ENVELOPE:
      return 0

    consumed_bytes = decoder.stream.tell()

    if self.version >= self._MIN_WIRE_FORMAT_VERSION:
      _, trailer_offset_tag = decoder.DecodeUint8()
      if trailer_offset_tag != definitions.BlinkSerializationTag.TRAILER_OFFSET:
        raise errors.ParserError("Trailer offset not found")
      _, self.trailer_offset = decoder.DecodeInt(
          byte_count=8, byte_order="big", signed=False
      )
      _, self.trailer_size = decoder.DecodeInt(
          byte_count=4, byte_order="big", signed=False
      )
      trailer_offset_data_size = (
          1 + 8 + 4
      )  # 1 + sizeof(uint64) + sizeof(uint32)
      consumed_bytes += trailer_offset_data_size
      if consumed_bytes >= len(self.raw_data):
        return 0
    return consumed_bytes

  def _ReadAESKey(self) -> tuple[definitions.WebCryptoKeyType, dict[str, Any]]:
    """Reads an AES CryptoKey.

    Returns:
      The AES key algorithm parameters
    """
    _, raw_id = self.deserializer.decoder.DecodeUint32Varint()
    crypto_key_algorithm_id = definitions.CryptoKeyAlgorithm(raw_id)

    _, length_bytes = self.deserializer.decoder.DecodeUint32Varint()

    algorithm_parameters = {
        "id": crypto_key_algorithm_id,
        "length_bits": length_bytes * 8,
    }

    return definitions.WebCryptoKeyType.SECRET, algorithm_parameters

  def _ReadHMACKey(self) -> tuple[definitions.WebCryptoKeyType, dict[str, Any]]:
    """Reads a HMAC CryptoKey.

    Returns:
      The HMAC key algorithm parameters
    """
    _, length_bytes = self.deserializer.decoder.DecodeUint32Varint()
    _, raw_hash = self.deserializer.decoder.DecodeUint32Varint()
    crypto_key_algorithm = definitions.CryptoKeyAlgorithm(raw_hash)

    algorithm_parameters = {
        "id": crypto_key_algorithm,
        "length_bits": length_bytes * 8,
    }

    return definitions.WebCryptoKeyType.SECRET, algorithm_parameters

  def _ReadRSAHashedKey(
      self,
  ) -> tuple[definitions.AsymmetricCryptoKeyType, dict[str, Any]]:
    """Reads an RSA Hashed CryptoKey.

    Returns:
      The RSA Hashed key algorithm parameters
    """
    _, raw_id = self.deserializer.decoder.DecodeUint32Varint()
    crypto_key_algorithm = definitions.CryptoKeyAlgorithm(raw_id)

    _, raw_key_type = self.deserializer.decoder.DecodeUint32Varint()
    key_type = definitions.AsymmetricCryptoKeyType(raw_key_type)

    _, modulus_length_bits = self.deserializer.decoder.DecodeUint32Varint()
    _, public_exponent_size = self.deserializer.decoder.DecodeUint32Varint()
    _, public_exponent_bytes = self.deserializer.decoder.ReadBytes(
        count=public_exponent_size
    )

    _, raw_hash = self.deserializer.decoder.DecodeUint32Varint()
    hash_algorithm = definitions.CryptoKeyAlgorithm(raw_hash)

    algorithm_parameters = {
        "id": crypto_key_algorithm,
        "modulus_length_bits": modulus_length_bits,
        "public_exponent_size": public_exponent_size,
        "public_exponent_bytes": public_exponent_bytes,
        "hash": hash_algorithm,
    }

    return key_type, algorithm_parameters

  def _ReadECKey(
      self,
  ) -> tuple[definitions.AsymmetricCryptoKeyType, dict[str, Any]]:
    """Reads an EC Key parameters from the current decoder position.

    Returns:
      The EC Key algorithm parameters.
    """
    _, raw_id = self.deserializer.decoder.DecodeUint32Varint()
    crypto_key_algorithm = definitions.CryptoKeyAlgorithm(raw_id)

    _, raw_key_type = self.deserializer.decoder.DecodeUint32Varint()
    key_type = definitions.AsymmetricCryptoKeyType(raw_key_type)

    _, raw_named_curve = self.deserializer.decoder.DecodeUint32Varint()
    named_curve_type = definitions.NamedCurve(raw_named_curve)

    algorithm_parameters = {
        "crypto_key_algorithm": crypto_key_algorithm,
        "named_curve_type": named_curve_type,
    }

    return key_type, algorithm_parameters

  def _ReadED25519Key(
      self,
  ) -> tuple[definitions.AsymmetricCryptoKeyType, dict[str, Any]]:
    """Reads a ED25519 CryptoKey from the current decoder position.

    Returns:
      The ED25519 key algorithm parameters.
    """
    _, raw_id = self.deserializer.decoder.DecodeUint32Varint()
    crypto_key_algorithm = definitions.CryptoKeyAlgorithm(raw_id)

    _, raw_key_type = self.deserializer.decoder.DecodeUint32Varint()
    key_type = definitions.AsymmetricCryptoKeyType(raw_key_type)

    algorithm_parameters = {"crypto_key_algorithm": crypto_key_algorithm}

    return key_type, algorithm_parameters

  def ReadNoParamsKey(
      self,
  ) -> tuple[definitions.WebCryptoKeyType, dict[str, Any]]:
    """Reads a No Params CryptoKey from the current decoder position.

    Returns:
      The No Parameters key algorithm parameters.
    """
    _, raw_id = self.deserializer.decoder.DecodeUint32Varint()
    crypto_key_algorithm = definitions.CryptoKeyAlgorithm(raw_id)

    algorithm_parameters = {"crypto_key_algorithm": crypto_key_algorithm}

    return definitions.WebCryptoKeyType.SECRET, algorithm_parameters

  def _ReadBlob(self) -> Optional[Blob]:
    """Reads a Blob from the current position.

    Returns:
      A Blob if the version is less then 3, None otherwise.
    """
    if self.version and self.version < 3:
      return None

    uuid = self.deserializer.ReadUTF8String()
    blob_type = self.deserializer.ReadUTF8String()
    _, size = self.deserializer.decoder.DecodeUint64Varint()
    return Blob(uuid=uuid, type=blob_type, size=size)

  def _ReadBlobIndex(self) -> Optional[BlobIndex]:
    """Reads a BlobIndex from the current decoder position.

    Returns:
      A parsed BlobIndex if the version is greater than 6, None otherwise.
    """
    if self.version < 6:
      return None
    _, index = self.deserializer.decoder.DecodeUint32Varint()
    return BlobIndex(index=index)

  def _ReadFile(self) -> Optional[File]:
    """Reads a Javascript File from the current position.

    Returns:
      A File instance, None if the version is less than 3.
    """
    if self.version < 3:
      return None

    path = self.deserializer.ReadUTF8String()
    name = self.deserializer.ReadUTF8String() if self.version >= 4 else None
    relative_path = (
        self.deserializer.ReadUTF8String() if self.version >= 4 else None
    )
    uuid = self.deserializer.ReadUTF8String()
    file_type = self.deserializer.ReadUTF8String()
    has_snapshot = (
        self.deserializer.decoder.DecodeUint32Varint()[1]
        if self.version >= 4
        else 0
    )

    if has_snapshot:
      _, size = self.deserializer.decoder.DecodeUint64Varint()
      _, last_modified_ms = self.deserializer.decoder.DecodeDouble()
      if self.version < 8:
        last_modified_ms *= _MS_PER_SECOND
    else:
      size = None
      last_modified_ms = None

    is_user_visible = (
        self.deserializer.decoder.DecodeUint32Varint()[1]
        if self.version >= 7
        else 1
    )

    return File(
        path=path,
        name=name,
        relative_path=relative_path,
        uuid=uuid,
        type=file_type,
        has_snapshot=has_snapshot,
        size=size,
        last_modified_ms=last_modified_ms,
        is_user_visible=is_user_visible,
    )

  def _ReadFileIndex(self) -> Optional[FileIndex]:
    """Reads a FileIndex from the current position.

    Returns:
      A FileIndex, or None if the version is less than 6.
    """
    if self.version < 6:
      return None
    _, index = self.deserializer.decoder.DecodeUint32Varint()
    return FileIndex(index=index)

  def _ReadFileList(self) -> Optional[FileList]:
    """Reads a Javascript FileList from the current position.

    Returns:
      A FileList, or None if a File could not be read.
    """
    _, length = self.deserializer.decoder.DecodeUint32Varint()
    files = []
    for _ in range(length):
      decoded_file = self._ReadFile()
      if not decoded_file:
        return None
      files.append(decoded_file)
    return FileList(files=files)

  def _ReadFileListIndex(self) -> Optional[FileListIndex]:
    """Reads a Javascript FileListIndex from the current position.

    Returns:
      A FileListIndex, or None if a FileIndex could not be read.
    """
    _, length = self.deserializer.decoder.DecodeUint32Varint()
    file_indexes = []
    for _ in range(length):
      decoded_file_index = self._ReadFileIndex()
      if not decoded_file_index:
        return None
      file_indexes.append(decoded_file_index)
    return FileListIndex(file_indexes=file_indexes)

  def _ReadImageBitmap(self) -> None:
    """Reads an ImageBitmap."""
    raise NotImplementedError("V8ScriptValueDecoder._ReadImageBitmap")

  def _ReadImageBitmapTransfer(self) -> ImageBitmapTransfer:
    """Reads an ImageBitmapTransfer."""
    _, index = self.deserializer.decoder.DecodeUint32Varint()
    return ImageBitmapTransfer(index=index)

  def _ReadImageData(self) -> None:
    """Reads an ImageData from the current position."""
    raise NotImplementedError("V8ScriptValueDecoder._ReadImageData")

  def _ReadDOMPoint(self) -> DOMPoint:
    """Reads a DOMPoint from the current position."""
    _, x = self.deserializer.decoder.DecodeDouble()
    _, y = self.deserializer.decoder.DecodeDouble()
    _, z = self.deserializer.decoder.DecodeDouble()
    _, w = self.deserializer.decoder.DecodeDouble()
    return DOMPoint(x=x, y=y, z=z, w=w)

  def _ReadDOMPointReadOnly(self) -> DOMPointReadOnly:
    """Reads a DOMPointReadOnly from the current position."""
    _, x = self.deserializer.decoder.DecodeDouble()
    _, y = self.deserializer.decoder.DecodeDouble()
    _, z = self.deserializer.decoder.DecodeDouble()
    _, w = self.deserializer.decoder.DecodeDouble()
    return DOMPointReadOnly(x=x, y=y, z=z, w=w)

  def _ReadDOMRect(self) -> DOMRect:
    """Reads a DOMRect from the current position."""
    _, x = self.deserializer.decoder.DecodeDouble()
    _, y = self.deserializer.decoder.DecodeDouble()
    _, width = self.deserializer.decoder.DecodeDouble()
    _, height = self.deserializer.decoder.DecodeDouble()
    return DOMRect(x=x, y=y, width=width, height=height)

  def _ReadDOMRectReadOnly(self) -> DOMRectReadOnly:
    """Reads a DOMRectReadOnly from the current position."""
    _, x = self.deserializer.decoder.DecodeDouble()
    _, y = self.deserializer.decoder.DecodeDouble()
    _, width = self.deserializer.decoder.DecodeDouble()
    _, height = self.deserializer.decoder.DecodeDouble()
    return DOMRectReadOnly(x=x, y=y, width=width, height=height)

  def _ReadDOMQuad(self) -> DOMQuad:
    """Reads a DOMQuad from the current position."""
    p1 = self._ReadDOMPoint()
    p2 = self._ReadDOMPoint()
    p3 = self._ReadDOMPoint()
    p4 = self._ReadDOMPoint()
    return DOMQuad(p1=p1, p2=p2, p3=p3, p4=p4)

  def _ReadDOMMatrix2D(self) -> DOMMatrix2D:
    """Reads a Javascript DOMMatrix2D from the current position."""
    values = [self.deserializer.decoder.DecodeDouble()[1] for _ in range(6)]
    return DOMMatrix2D(values=values)

  def _ReadDOMMatrix2DReadOnly(self) -> DOMMatrix2DReadOnly:
    """Reads a Javascript Read-Only DOMMatrix2D from the current position."""
    values = [self.deserializer.decoder.DecodeDouble()[1] for _ in range(6)]
    return DOMMatrix2DReadOnly(values=values)

  def _ReadDOMMatrix(self) -> DOMMatrix:
    """Reads a Javascript DOMMatrix from the current position."""
    values = [self.deserializer.decoder.DecodeDouble()[1] for _ in range(16)]
    return DOMMatrix(values=values)

  def _ReadDOMMatrixReadOnly(self) -> DOMMatrixReadOnly:
    """Reads a Javascript Read-Only DOMMatrix from the current position."""
    values = [self.deserializer.decoder.DecodeDouble()[1] for _ in range(16)]
    return DOMMatrixReadOnly(values=values)

  def _ReadMessagePort(self) -> MessagePort:
    """Reads a MessagePort from the current position."""
    _, index = self.deserializer.decoder.DecodeUint32Varint()
    return MessagePort(index=index)

  def _ReadMojoHandle(self) -> MojoHandle:
    """Reads a MojoHandle from the current position."""
    _, index = self.deserializer.decoder.DecodeUint32Varint()
    return MojoHandle(index=index)

  def _ReadOffscreenCanvasTransfer(self) -> OffscreenCanvasTransfer:
    """Reads a Offscreen Canvas Transfer from the current position."""
    _, width = self.deserializer.decoder.DecodeUint32Varint()
    _, height = self.deserializer.decoder.DecodeUint32Varint()
    _, canvas_id = self.deserializer.decoder.DecodeUint32Varint()
    _, client_id = self.deserializer.decoder.DecodeUint32Varint()
    _, sink_id = self.deserializer.decoder.DecodeUint32Varint()
    _, filter_quality = self.deserializer.decoder.DecodeUint32Varint()

    return OffscreenCanvasTransfer(
        width=width,
        height=height,
        canvas_id=canvas_id,
        client_id=client_id,
        sink_id=sink_id,
        filter_quality=filter_quality,
    )

  def _ReadReadableStreamTransfer(self) -> ReadableStreamTransfer:
    """Reads a Readable Stream Transfer from the current position."""
    _, index = self.deserializer.decoder.DecodeUint32Varint()
    return ReadableStreamTransfer(index=index)

  def _ReadWriteableStreamTransfer(self) -> WriteableStreamTransfer:
    """Reads a Writeable Stream Transfer from the current position."""
    _, index = self.deserializer.decoder.DecodeUint32Varint()
    return WriteableStreamTransfer(index=index)

  def _ReadTransformStreamTransfer(self) -> TransformStreamTransfer:
    """Reads a TransformStreamTransfer from the current position."""
    _, index = self.deserializer.decoder.DecodeUint32Varint()
    return TransformStreamTransfer(index=index)

  def _ReadDOMException(self) -> DOMException:
    """Reads a DOMException from the current position."""
    name = self.deserializer.ReadUTF8String()
    message = self.deserializer.ReadUTF8String()
    stack_unused = self.deserializer.ReadUTF8String()
    return DOMException(name=name, message=message, stack_unused=stack_unused)

  def _ReadRTCEncodedAudioFrame(self) -> RTCEncodedAudioFrame:
    """Reads a RTC Encoded Audio Frame from the current position."""
    _, index = self.deserializer.decoder.DecodeUint32Varint()
    return RTCEncodedAudioFrame(index=index)

  def _ReadRTCEncodedVideoFrame(self) -> RTCEncodedVideoFrame:
    """Reads a RTC Encoded Video Frame from the current position."""
    _, index = self.deserializer.decoder.DecodeUint32Varint()
    return RTCEncodedVideoFrame(index=index)

  def _ReadCryptoKey(self) -> CryptoKey:
    """Reads a CryptoKey from the current position.

    Returns:
      A parsed CryptoKey.

    Raises:
      ParserError: if there is an unexpected CryptoKeySubTag.
    """
    _, raw_key_byte = self.deserializer.decoder.DecodeUint8()
    key_byte = definitions.CryptoKeySubTag(raw_key_byte)
    algorithm_parameters: dict[str, Any] = {}
    key_type: Union[
        definitions.WebCryptoKeyType, definitions.AsymmetricCryptoKeyType, None
    ] = None
    if key_byte == definitions.CryptoKeySubTag.AES_KEY:
      key_type, algorithm_parameters = self._ReadAESKey()
    elif key_byte == definitions.CryptoKeySubTag.HMAC_KEY:
      key_type, algorithm_parameters = self._ReadHMACKey()
    elif key_byte == definitions.CryptoKeySubTag.RSA_HASHED_KEY:
      key_type, algorithm_parameters = self._ReadRSAHashedKey()
    elif key_byte == definitions.CryptoKeySubTag.EC_KEY:
      key_type, algorithm_parameters = self._ReadECKey()
    elif key_byte == definitions.CryptoKeySubTag.ED25519_KEY:
      key_type, algorithm_parameters = self._ReadED25519Key()
    elif key_byte == definitions.CryptoKeySubTag.NO_PARAMS_KEY:
      key_type, algorithm_parameters = self.ReadNoParamsKey()
    else:
      raise errors.ParserError("Unexpected CryptoKeySubTag")

    _, raw_usages = self.deserializer.decoder.DecodeUint32Varint()
    usages = definitions.CryptoKeyUsage(raw_usages)

    extractable = bool(raw_usages & definitions.CryptoKeyUsage.EXTRACTABLE)
    _, key_data_length = self.deserializer.decoder.DecodeUint32Varint()

    _, key_data = self.deserializer.decoder.ReadBytes(count=key_data_length)

    return CryptoKey(
        key_type=key_type,
        algorithm_parameters=algorithm_parameters,
        extractable=extractable,
        usages=usages,
        key_data=key_data,
    )

  def _ReadAudioData(self) -> AudioData:
    """Reads an AudioData from the current position."""
    _, audio_frame_index = self.deserializer.decoder.DecodeUint32Varint()
    return AudioData(audio_frame_index=audio_frame_index)

  def _ReadDomFileSystem(self) -> DOMFileSystem:
    """Reads an DOMFileSystem from the current position."""
    _, raw_type = self.deserializer.decoder.DecodeUint32Varint()
    name = self.deserializer.ReadUTF8String()
    root_url = self.deserializer.ReadUTF8String()
    return DOMFileSystem(raw_type=raw_type, name=name, root_url=root_url)

  def _ReadFileSystemFileHandle(self) -> FileSystemHandle:
    """Reads a FileSystemHandle from the current position."""
    name = self.deserializer.ReadUTF8String()
    _, token_index = self.deserializer.decoder.DecodeUint32Varint()
    return FileSystemHandle(name=name, token_index=token_index)

  def _ReadVideoFrame(self) -> VideoFrame:
    """Reads the video frame from the current position."""
    _, index = self.deserializer.decoder.DecodeUint32Varint()
    return VideoFrame(index=index)

  def _ReadEncodedAudioChunk(self) -> EncodedAudioChunk:
    """Reads the encoded audio chunk from the current position."""
    _, index = self.deserializer.decoder.DecodeUint32Varint()
    return EncodedAudioChunk(index=index)

  def _ReadEncodedVideoChunk(self) -> EncodedVideoChunk:
    """Reads the encoded video chunk from the current position."""
    _, index = self.deserializer.decoder.DecodeUint32Varint()
    return EncodedVideoChunk(index=index)

  def _ReadMediaStreamTrack(self) -> None:
    """Reads the media stream track from the current position."""
    raise NotImplementedError("V8ScriptValueDecoder._ReadMediaStreamTrack")

  def _ReadCropTarget(self) -> None:
    """Reads the crop target from the current position."""
    raise NotImplementedError("V8ScriptValueDecoder._ReadCropTarget")

  def _ReadRestrictionTarget(self) -> None:
    """Reads the restriction target from the current position."""
    raise NotImplementedError("V8ScriptValueDecoder._ReadRestrictionTarget")

  def _ReadMediaSourceHandle(self) -> MediaSourceHandle:
    """Reads the media source handle from the current position."""
    _, index = self.deserializer.decoder.DecodeUint32Varint()
    return MediaSourceHandle(index=index)

  def _ReadFencedFrameConfig(self) -> None:
    """Reads the fenced frame target from the current position."""
    raise NotImplementedError("V8ScriptValueDecoder._ReadFencedFrameConfig")

  def ReadTag(self) -> definitions.BlinkSerializationTag:
    """Reads a blink serialization tag from the current position.

    Returns:
      The blink serialization tag.

    Raises:
      ParserError: if an invalid blink tag is read.
    """
    _, tag_value = self.deserializer.decoder.DecodeUint8()
    try:
      return definitions.BlinkSerializationTag(tag_value)
    except ValueError as err:
      offset = self.deserializer.decoder.stream.tell()
      raise errors.ParserError(
          f"Invalid blink tag encountered at offset {offset}"
      ) from err

  def ReadHostObject(self) -> Any:
    """Reads a host object from the current position.

    Returns:
      The Host Object.
    """
    tag = self.ReadTag()
    dom_object: Any = None

    # V8ScriptValueDeserializer
    if tag == definitions.BlinkSerializationTag.BLOB:
      dom_object = self._ReadBlob()
    elif tag == definitions.BlinkSerializationTag.BLOB_INDEX:
      dom_object = self._ReadBlobIndex()
    elif tag == definitions.BlinkSerializationTag.FILE:
      dom_object = self._ReadFile()
    elif tag == definitions.BlinkSerializationTag.FILE_INDEX:
      dom_object = self._ReadFileIndex()
    elif tag == definitions.BlinkSerializationTag.FILE_LIST:
      dom_object = self._ReadFileList()
    elif tag == definitions.BlinkSerializationTag.FILE_LIST_INDEX:
      dom_object = self._ReadFileListIndex()
    elif tag == definitions.BlinkSerializationTag.IMAGE_BITMAP:
      self._ReadImageBitmap()
    elif tag == definitions.BlinkSerializationTag.IMAGE_BITMAP_TRANSFER:
      dom_object = self._ReadImageBitmapTransfer()
    elif tag == definitions.BlinkSerializationTag.IMAGE_DATA:
      self._ReadImageData()
    elif tag == definitions.BlinkSerializationTag.DOM_POINT:
      dom_object = self._ReadDOMPoint()
    elif tag == definitions.BlinkSerializationTag.DOM_POINT_READ_ONLY:
      dom_object = self._ReadDOMPointReadOnly()
    elif tag == definitions.BlinkSerializationTag.DOM_RECT:
      dom_object = self._ReadDOMRect()
    elif tag == definitions.BlinkSerializationTag.DOM_RECT_READ_ONLY:
      dom_object = self._ReadDOMRectReadOnly()
    elif tag == definitions.BlinkSerializationTag.DOM_QUAD:
      dom_object = self._ReadDOMQuad()
    elif tag == definitions.BlinkSerializationTag.DOM_MATRIX2D:
      dom_object = self._ReadDOMMatrix2D()
    elif tag == definitions.BlinkSerializationTag.DOM_MATRIX2D_READ_ONLY:
      dom_object = self._ReadDOMMatrix2DReadOnly()
    elif tag == definitions.BlinkSerializationTag.DOM_MATRIX:
      dom_object = self._ReadDOMMatrix()
    elif tag == definitions.BlinkSerializationTag.DOM_MATRIX_READ_ONLY:
      dom_object = self._ReadDOMMatrixReadOnly()
    elif tag == definitions.BlinkSerializationTag.MESSAGE_PORT:
      dom_object = self._ReadMessagePort()
    elif tag == definitions.BlinkSerializationTag.MOJO_HANDLE:
      dom_object = self._ReadMojoHandle()
    elif tag == definitions.BlinkSerializationTag.OFFSCREEN_CANVAS_TRANSFER:
      dom_object = self._ReadOffscreenCanvasTransfer()
    elif tag == definitions.BlinkSerializationTag.READABLE_STREAM_TRANSFER:
      dom_object = self._ReadReadableStreamTransfer()
    elif tag == definitions.BlinkSerializationTag.WRITABLE_STREAM_TRANSFER:
      dom_object = self._ReadWriteableStreamTransfer()
    elif tag == definitions.BlinkSerializationTag.TRANSFORM_STREAM_TRANSFER:
      dom_object = self._ReadTransformStreamTransfer()
    elif tag == definitions.BlinkSerializationTag.DOM_EXCEPTION:
      dom_object = self._ReadDOMException()

    # V8ScriptValueDeserializerForModules
    elif tag == definitions.BlinkSerializationTag.CRYPTO_KEY:
      dom_object = self._ReadCryptoKey()
    elif tag == definitions.BlinkSerializationTag.DOM_FILE_SYSTEM:
      dom_object = self._ReadDomFileSystem()
    elif tag == definitions.BlinkSerializationTag.FILE_SYSTEM_FILE_HANDLE:
      dom_object = self._ReadFileSystemFileHandle()
    elif tag == definitions.BlinkSerializationTag.RTC_ENCODED_AUDIO_FRAME:
      dom_object = self._ReadRTCEncodedAudioFrame()
    elif tag == definitions.BlinkSerializationTag.RTC_ENCODED_VIDEO_FRAME:
      dom_object = self._ReadRTCEncodedVideoFrame()
    elif tag == definitions.BlinkSerializationTag.AUDIO_DATA:
      dom_object = self._ReadAudioData()
    elif tag == definitions.BlinkSerializationTag.VIDEO_FRAME:
      dom_object = self._ReadVideoFrame()
    elif tag == definitions.BlinkSerializationTag.ENCODED_AUDIO_CHUNK:
      dom_object = self._ReadEncodedAudioChunk()
    elif tag == definitions.BlinkSerializationTag.ENCODED_VIDEO_CHUNK:
      dom_object = self._ReadEncodedVideoChunk()
    elif tag == definitions.BlinkSerializationTag.MEDIA_STREAM_TRACK:
      self._ReadMediaStreamTrack()
    elif tag == definitions.BlinkSerializationTag.CROP_TARGET:
      self._ReadCropTarget()
    elif tag == definitions.BlinkSerializationTag.RESTRICTION_TARGET:
      self._ReadRestrictionTarget()
    elif tag == definitions.BlinkSerializationTag.MEDIA_SOURCE_HANDLE:
      dom_object = self._ReadMediaSourceHandle()
    elif tag == definitions.BlinkSerializationTag.FENCED_FRAME_CONFIG:
      self._ReadFencedFrameConfig()
    return dom_object

  def Deserialize(self) -> Any:
    """Deserializes a Blink SSV.

    The serialization format has two 'envelopes'.
    [version tag] [Blink version] [version tag] [v8 version] ...

    Returns:
      The deserialized script value.

    Raises:
      ParserError: if an unsupported header version was found.
    """
    version_envelope_size = self._ReadVersionEnvelope()
    if self.trailer_size:
      self.deserializer = v8.ValueDeserializer(
          io.BytesIO(
              self.raw_data[version_envelope_size : self.trailer_offset]
          ),
          delegate=self,
      )
    else:
      self.deserializer = v8.ValueDeserializer(
          io.BytesIO(self.raw_data[version_envelope_size:]), delegate=self
      )
    is_supported = self.deserializer.ReadHeader()
    if not is_supported:
      raise errors.ParserError("Unsupported header")
    return self.deserializer.ReadValue()

  @classmethod
  def FromBytes(cls, data: bytes) -> Any:
    """Deserializes a Blink SSV from the data array.

    Returns:
      The deserialized script value.
    """
    if len(data) < 3:
      raise errors.ParserError("Data too short to contain a Blink SSV")
    if (
        data[0] == definitions.BlinkSerializationTag.VERSION
        and data[1] == definitions.REQUIRES_PROCESSING_SSV_PSEUDO_VERSION
        and data[2] == definitions.COMPRESSED_WITH_SNAPPY
    ):
      # ignore the wrapped header bytes when decompressing
      data = snappy.decompress(data[3:])
    return cls(data).Deserialize()
